import logging

from m.color import color
from m.core import Issue, OneOf, hone
from m.core import io as mio
from m.core import one_of
from m.log.misc import default_record_fmt, format_context, format_location
from pydantic import BaseModel, Field

from ..types import EnvVars, Message, ProviderModule


class GithubEnvVars(BaseModel):
    """Environment variables required when running in Github."""

    repo: str = Field('GITHUB_REPOSITORY')
    run_id: str = Field('GITHUB_RUN_ID')
    run_number: str = Field('GITHUB_RUN_NUMBER')
    github_token: str = Field('GITHUB_TOKEN')
    git_branch: str = Field('GITHUB_REF')
    git_sha: str = Field('GITHUB_SHA')
    triggered_by: str = Field('GITHUB_ACTOR')


def env_vars() -> OneOf[Issue, EnvVars]:
    """Read the environment variables from Github Actions.

    Returns:
        An `EnvVars` instance or an Issue.
    """
    server_url = 'https://github.com'
    return one_of(
        lambda: [
            EnvVars(
                **env.model_dump(),
                ci_env=True,
                server_url=server_url,
                run_url=run_url,
            )
            for env in mio.env_model(GithubEnvVars)
            for run_url in (
                f'{server_url}/{env.repo}/actions/runs/{env.run_id}',
            )
        ],
    ).flat_map_bad(hone('GH Actions env_vars failure'))


def _gh_format(msg_type: str, msg: Message, message: str, postfix: str) -> str:
    """Format an error or warning message for Github.

    See::

        https://docs.github.com/en/actions/reference/workflow-commands-for-github-actions#setting-an-error-message
        https://docs.github.com/en/actions/reference/workflow-commands-for-github-actions#setting-a-warning-message

    Args:
        msg_type: either error or warnings.
        msg: A Message instance.
        message: A simple message.
        postfix: A string to print after the simple message.

    Returns:
        A formatted string.
    """
    file_entry = f'file={msg.file}' if msg.file else ''
    line_entry = f'line={msg.line}' if msg.line else ''
    col_entry = f'col={msg.col}' if msg.col else ''
    parts = [x for x in (file_entry, line_entry, col_entry) if x]
    loc = ','.join(parts)
    msg_info = f' {loc}' if loc else ''
    return f'::{msg_type}{msg_info}::{message}{postfix}'


def _format_block(name, desc) -> str:
    with_desc = f'{{gray}} - {desc}' if desc else ''
    return color(
        f'::group::{{bold_purple}}{name}',
        with_desc,
    )


def log_format(
    formatter: logging.Formatter,
    record: logging.LogRecord,
    show_traceback: bool,
    debug_python: bool,
) -> str:
    """Format a log record using the functions provided in this module.

    This function is meant to be used by a log formatter. See more info


    Args:
        formatter: A log formatter instance.
        record: A log record.
        show_traceback: If true, display the python stack trace.
        debug_python: If true, display the location of the record's origin.

    Returns:
        A formatted string.
    """
    record_dict = record.__dict__

    open_b = record_dict.get('open_block')
    if open_b:
        return _format_block(*open_b)

    if record_dict.get('close_block'):
        return '::endgroup::'

    is_command = record.levelname in {'WARNING', 'ERROR'}

    indent_padding = 2 if is_command else 3
    indent = len(record.levelname) + indent_padding

    ci_info = record_dict.get('ci_info', Message(msg=record.msg))
    context = format_context(record, indent, show_traceback=show_traceback)
    if not record.msg:
        return context[1:]

    loc = (
        format_location([record.pathname, f'{record.lineno}'])
        if debug_python
        else ''
    )
    if is_command:
        msg = f'{loc} {record.msg}'.lstrip()
        return _gh_format(record.levelname.lower(), ci_info, msg, context)
    return default_record_fmt(
        record,
        formatter.formatTime(record, formatter.datefmt),
        loc,
        context,
    )


tool = ProviderModule(
    ci=True,
    env_vars=env_vars,
    formatter=log_format,
)



import json
from pathlib import Path
from typing import Any

from m.core import rw as mio
from m.log import EnvVars, get_ci_tool
from pydantic import BaseModel

from ..core import Res, fp, hone, one_of
from .config import Config, read_config
from .git_env import GitEnv, get_git_env
from .release_env import ReleaseEnv, get_release_env


class MEnv(BaseModel):
    """Contains all the information required for a CI run."""

    config: Config
    env_vars: EnvVars
    git_env: GitEnv
    release_env: ReleaseEnv


def get_m_env(m_dir: str) -> Res[MEnv]:
    """Obtain the M Environment object.

    Args:
        m_dir: The directory containing the m configuration.

    Returns:
        The M Environment if it exists otherwise an issue.
    """
    ci_tool = get_ci_tool()
    return one_of(lambda: [
        MEnv(
            config=config,
            env_vars=env_vars,
            git_env=git_env,
            release_env=release_env,
        )
        for config in read_config(m_dir)
        for env_vars in ci_tool.env_vars()
        for git_env in get_git_env(config, env_vars)
        for release_env in get_release_env(config, env_vars, git_env)
    ]).flat_map_bad(hone('get_m_env failure'))


def _m_env_vars(m_env: MEnv) -> Res[list[str]]:
    """Serialize the m environment variables.

    Args:
        m_env: The `M` environment.

    Returns:
        A string if successful.
    """
    config = m_env.config
    env_vars = m_env.env_vars
    git = m_env.git_env
    release = m_env.release_env
    associated_pr = (
        git.pull_request or (git.commit and git.commit.associated_pull_request)
    )
    associated_pr_num = associated_pr.pr_number if associated_pr else ''
    cache_from_pr = git.get_pr_number() or associated_pr_num
    env = {
        'M_DIR': config.m_dir,
        'M_OWNER': config.owner,
        'M_REPO': config.repo,
        'M_CI': env_vars.ci_env,
        'M_WORKFLOW': config.workflow,
        'M_RUN_ID': env_vars.run_id,
        'M_RUN_NUMBER': env_vars.run_number,
        'M_SHA': git.sha,
        'M_BRANCH': git.branch,
        'M_TARGET_BRANCH': git.target_branch,
        'M_ASSOCIATED_PR_NUMBER': associated_pr_num,
        'M_PR_BRANCH': git.get_pr_branch(),
        'M_PR_NUMBER': git.get_pr_number(),
        'M_TAG': release.build_tag,
        'M_PY_TAG': release.python_tag,
        'M_CACHE_FROM_PR': cache_from_pr,
        'M_IS_RELEASE': release.is_release,
        'M_IS_RELEASE_PR': release.is_release_pr,
        'M_IS_HOTFIX_PR': release.is_hotfix_pr,
    }
    lines = [f'{env_key}={env_val}' for env_key, env_val in env.items()]
    return fp.Good(lines)


def write_m_env_vars(m_dir: str) -> Res[Any]:
    """Write a file with the M environment variables.

    Args:
        m_dir: The directory with the m configuration.

    Returns:
        An issue or the m environment instance.
    """
    target_dir = Path(f'{m_dir}/.m')

    if not Path.exists(target_dir):
        Path.mkdir(target_dir, parents=True)
    return one_of(lambda: [
        json.loads(m_env.model_dump_json())
        for m_env in get_m_env(m_dir)
        for env_list in _m_env_vars(m_env)
        for _ in mio.write_file(f'{m_dir}/.m/env.list', '\n'.join(env_list))
    ])


def _lower_bool(line: str) -> str:
    return line.replace('=False', '=false').replace('=True', '=true')


def bashrc_snippet(m_dir: str) -> Res[str]:
    """Create a bash snippet that exports the M environment variables.

    Args:
        m_dir: The directory with the m configuration.

    Returns:
        An issue or a bash snippet.
    """
    return one_of(lambda: [
        '\n'.join([
            f'export {assignment}'
            for line in env_list
            for assignment in fp.Good(_lower_bool(line))
        ])
        for m_env in get_m_env(m_dir)
        for env_list in _m_env_vars(m_env)
    ])



from textwrap import dedent
from typing import Any

from m.core import yaml
from pydantic import BaseModel

from .image import DockerImage

TEMPLATE = """\
# AUTOGENERATED FILE - Update {m_dir}/m.yaml
#   then run `m blueprints --update-workflow` to update
name: m

on:
  workflow_call:
    inputs:
      m-tag:
        type: string
        description: The unique version to use for all the images
        required: true
      cache-from-pr:
        type: string
        description: The pull request number to attempt to use as cache.
        required: true

permissions: write-all

env:{global_env}

jobs:
  blueprints:
    runs-on: {default_runner}
    steps:
      - name: checkout
        uses: actions/checkout@v4
      - name: install-m
        run: pip install jmlopez-m
      - name: m-blueprints
        id: m-blueprints
        run: |-
          m blueprints --skip-makefile --skip-workflow {m_dir}
      - name: archive
        uses: actions/upload-artifact@v3
        with:
          name: m-blueprints
          path: {ci_dir}
      - name: chmod
        run: chmod +x {ci_dir}/*.sh
      - {docker_login}
      {build_steps}
"""


class TemplateVars(BaseModel):
    """Template variables."""

    m_dir: str

    ci_dir: str

    default_runner: str

    global_env: str

    docker_login: str

    build_steps: str


def _indent(text: str, num: int) -> str:
    spaces = '  ' * num
    return text.replace('\n', f'\n{spaces}', -1)


class Workflow(BaseModel):
    """Helper class to write the `m` workflow."""

    m_dir: str

    ci_dir: str

    global_env: dict[str, str] | None

    default_runner: str

    docker_registry: str

    extra_build_steps: list[dict[str, Any]] | None

    images: list[DockerImage]

    def docker_login_str(self: 'Workflow') -> str:
        """Generate a github action str to login to docker.

        Returns:
            A string to add to the Github workflow.
        """
        login_obj = """\
            name: docker-login
            uses: docker/login-action@v3
            with:
              registry: ghcr.io
              username: ${{ github.actor }}
              password: ${{ secrets.GITHUB_TOKEN }}"""
        return _indent(dedent(login_obj), 4)

    def global_env_str(self: 'Workflow') -> str:
        """Generate a github action str with the global environment variables.

        Returns:
            A string to add to the Github workflow.
        """
        all_vars = {
            'GITHUB_TOKEN': '${{ secrets.GITHUB_TOKEN }}',
            'M_TAG': '${{ inputs.m-tag }}',
            'M_CACHE_FROM_PR': '${{ inputs.cache-from-pr }}',
            **(self.global_env or {}),
        }
        vars_str = '\n'.join([
            f'  {env_var}: {env_val}'
            for env_var, env_val in all_vars.items()
        ])
        return f'\n{vars_str}'

    def build_steps_str(self: 'Workflow') -> str:
        """Generate a github action str for the build steps.

        Returns:
            A string to add to the Github workflow.
        """
        lines: list[str] = []
        for step in self.extra_build_steps or []:
            step_str = _indent(yaml.dumps(step)[:-1], 1)
            lines.append(f'- {step_str}')
        for img in self.images:
            cache_sh = f'{self.ci_dir}/_find-cache.sh'
            build_sh = f'{self.ci_dir}/{img.image_name}.build.sh'
            push_sh = f'{self.ci_dir}/_push-image.sh'
            image_steps = [
                f'- name: {img.image_name} - cache',
                f'  run: {cache_sh} {img.image_name}',
                f'- name: {img.image_name} - build',
                f'  run: {build_sh}',
                f'- name: {img.image_name} - push',
                f'  run: {push_sh} {img.image_name}',
            ]
            lines.extend(image_steps)
        return _indent('\n'.join(lines), 3)

    def __str__(self: 'Workflow') -> str:
        """Stringify the workflow file.

        Returns:
            The github workflow.
        """
        template_vars = TemplateVars(
            m_dir=self.m_dir,
            default_runner=self.default_runner,
            global_env=self.global_env_str(),
            ci_dir=self.ci_dir,
            docker_login=self.docker_login_str(),
            build_steps=self.build_steps_str(),
        )
        return TEMPLATE.format(**template_vars.model_dump())



from m.cli import Arg, BaseModel, command, run_main
from m.core.io import env


class Arguments(BaseModel):
    r"""Merge a pull request.

    https://docs.github.com/en/rest/reference/pulls#merge-a-pull-request

    example::

        $ m github merge_pr \\
            --owner owner \\
            --repo repo \\
            --commit-title 'commit_title' \\
            99
    """

    owner: str = Arg(
        default=env('GITHUB_REPOSITORY_OWNER'),
        help='repo owner',
    )
    repo: str = Arg(
        help='repo name',
        required=True,
    )
    commit_title: str | None = Arg(
        help='commit title',
    )
    pr: int = Arg(
        help='the pr number',
        positional=True,
        required=True,
    )


@command(
    help='merge a pull request',
    model=Arguments,
)
def run(arg: Arguments, arg_ns) -> int:
    from m.github.api import merge_pr
    return run_main(lambda: merge_pr(
        arg_ns.token,
        arg.owner,
        arg.repo,
        arg.pr,
        arg.commit_title,
    ))



# noqa: WPS412
from .colors import color
from .disable import color_disabled
from .pygment import highlight_json, highlight_yaml

__all__ = [  # noqa: WPS410
    'color_disabled',
    'highlight_json',
    'highlight_yaml',
    'color',
]



import json
import logging
import textwrap
from typing import cast

from m.color import color, highlight_json, highlight_yaml
from m.core import Issue, yaml

from .logger import PROMPT


def indent_payload(
    indent: int,
    payload: dict,
    prepend_new_line: bool = True,
    as_yaml: bool = True,
) -> str:
    """Stringify a dictionary as JSON and indent it.

    Args:
        indent: The number of spaces to indent.
        payload: The data to stringify and indent.
        prepend_new_line: Prepend a new line to the payload if `True`.
        as_yaml: Dump the payload in yaml.

    Returns:
        An indented payload.
    """
    json_dict = (
        highlight_yaml(yaml.dumps(payload))
        if as_yaml
        else highlight_json(json.dumps(payload, indent=2))
    )
    indented_payload = textwrap.indent(json_dict, ' ' * indent)
    return f'\n{indented_payload}' if prepend_new_line else indented_payload


def format_context(
    record: logging.LogRecord,
    indent: int,
    show_traceback: bool,
) -> str:
    """Extract the context from a log record.

    Args:
        record: The log record instance.
        indent: The level of indentation to apply to the string.
        show_traceback: Attempts to display the traceback if it exists.

    Returns:
        A string with leading new line of the context associated with a log
        record.
    """
    context = record.__dict__.get('context')
    data_dict = None
    as_yaml = Issue.yaml_traceback
    if isinstance(context, Issue):
        data_dict = cast(dict, context.to_dict(show_traceback=show_traceback))
        if context.only_context():
            data_dict = data_dict['context']
    else:
        data_dict = context
    return (
        indent_payload(indent, data_dict, as_yaml=as_yaml)
        if data_dict
        else ''
    )


def format_location(
    parts: list[str | None],
    default_color: str = 'aqua',
) -> str:
    """Join a list of possible defined strings.

    Args:
        parts: A list of possibly defined strings.
        default_color: The color to use for the string.

    Returns:
        A string with the location or an empty string.
    """
    loc = ':'.join([x for x in parts if x])
    return color(f'{{{default_color}}}[{loc}]{{end}}') if loc else ''


def default_record_fmt(
    record: logging.LogRecord,
    asctime: str,
    after_time: str,
    after_msg: str,
) -> str:
    """Format a log record.

    Args:
        record: A log record.
        asctime: The string representing the time of creation.
        after_time: A string that will be placed after the time.
        after_msg: A string to be placed after the main message.

    Returns:
        A formatted string.
    """
    level_color = {
        logging.DEBUG: 'green',
        logging.INFO: 'blue',
        logging.WARNING: 'yellow',
        PROMPT: 'bold_blue',
        logging.ERROR: 'red',
        logging.CRITICAL: 'bold_red',
    }.get(record.levelno, 'gray')
    fmt_time = color(f'{{gray}}[{asctime}]{{end}}') if asctime else ''
    fmt_level = color(f'{{{level_color}}}[{record.levelname}]{{end}}')
    prefix = f'{fmt_level} {fmt_time}{after_time}'.rstrip()
    record_msg = record.msg
    if level_color != 'blue':
        record_msg = color(f'{{{level_color}}}{record_msg}{{end}}')
    return f'{prefix}: {record_msg}{after_msg}'



from typing import cast

from ..types import AnyMap, FuncArgs


def should_handle(field: AnyMap) -> bool:
    """Handle the proxy field.

    Args:
        field: A dictionary with information for a cli argument.

    Returns:
        True if it should handle the field as a proxy to the parser.
    """
    return field.get('proxy', None) is not None


def handle_field(field: AnyMap) -> FuncArgs:
    """Pass the `proxy` property in the field as is.

    Args:
        field: A dictionary representation of the field.

    Returns:
        Function arguments for the parser `add_argument` method.
    """
    # relying on documentation to state that `proxy` should be used to provide
    # the arguments.
    return cast(FuncArgs, field['proxy'])



from inspect import signature
from typing import Any, Callable

from .types import MISSING


def params_count(func: Callable) -> int:
    """Compute the number of parameters in a function.

    Args:
        func: The function in question.

    Returns:
        the number of parameters.
    """
    sig = signature(func)
    return len(sig.parameters)


def argument_name(name: str) -> str:
    """Normalize an argument name.

    Args:
        name: Name of the argument.

    Returns:
        Normalized name of the argument.
    """
    cli_arg_name = name.replace('_', '-')
    dashes = '-' if len(cli_arg_name) == 1 else '--'
    return f'{dashes}{cli_arg_name}'


def argument_description(
    description: str,
    default: Any | None = MISSING,
) -> str:
    """Append default value to argument description.

    Args:
        description: argument description.
        default: Argument's default value.

    Returns:
        The description of the argument.
    """
    default = f'(default: {default})' if default is not MISSING else ''
    return f'{description} {default}'.strip()



from pathlib import Path
from typing import Any, Callable

import yaml

from . import rw
from .issue import Issue
from .json import parse_json
from .one_of import Good, OneOf, hone, issue, one_of

ParserFunction = Callable[[str, bool], OneOf[Issue, Any]]


def parse_yaml(
    yaml_str: str,
    error_if_empty: bool = False,
) -> OneOf[Issue, Any]:
    """Parse a string as yaml.

    Args:
        yaml_str: The string to parse.
        error_if_empty: The parser may throw an error if `True`.

    Returns:
        A `Good` containing the parsed contents of the yaml string.
    """
    empty = '' if error_if_empty else 'null'
    try:
        return Good(yaml.safe_load(yaml_str or empty))
    except Exception as ex:
        return issue('failed to parse the yaml data', cause=ex)


def read_yson(
    filename: str,
    error_if_empty: bool = False,
) -> OneOf[Issue, Any]:
    """Read a json object from a json or yaml file.

    It will choose the parser based on the filename extension.

    Args:
        filename: The filename to read from, if `None` it reads from stdin.
        error_if_empty: The json parser may throw an error if `True`.

    Returns:
        A `Good` containing the parsed contents of the json file.
    """
    empty: str = '' if error_if_empty else 'null'
    file_map: dict[str, ParserFunction] = {
        '.yaml': parse_yaml,
        '.yml': parse_yaml,
        '.json': parse_json,
    }
    ext = Path(filename).suffix
    parser = file_map.get(ext, parse_json)
    context = {'filename': filename or 'SYS.STDIN'}
    return one_of(lambda: [
        json_data
        for json_str in rw.read_file(filename)
        for json_data in parser(json_str or empty, error_if_empty)
    ]).flat_map_bad(hone(f'failed to read "{ext}" file', context=context))



import argparse

from m.cli import Meta

meta = Meta(
    help='execute commands',
    description="""
        See the help for each of the following supported commands.
    """,
)


def add_arguments(parser: argparse.ArgumentParser) -> None:
    """Handle an ArgumentParser instance to add global cli arguments.

    Args:
        parser: The Argument parser instance.
    """
    from m.version import VERSION

    parser.add_argument('--version', action='version', version=VERSION)



from m.core import Bad, Good, Res, hone, one_of, subprocess


def get_branch() -> Res[str]:
    """Get the current git branch name.

    Returns:
        A `OneOf` containing an `Issue` or a string specifying the branch.
    """
    return subprocess.eval_cmd('git rev-parse --abbrev-ref HEAD')


def stage_all() -> Res[str]:
    """Stage the current changes in the branch.

    Returns:
        A `OneOf` containing an `Issue` or the response from `git add .`.
    """
    res = subprocess.eval_cmd('git add .')
    return res.flat_map_bad(hone('git add failure'))


def commit(msg: str) -> Res[str]:
    """Create a commit.

    Args:
        msg: commit description.

    Returns:
        A `OneOf` containing an `Issue` or the response from the git command.
    """
    res = subprocess.eval_cmd(f'git commit -m "{msg}"')
    return res.flat_map_bad(hone('git commit failure'))


def push_branch(branch: str) -> Res[str]:
    """Push branch.

    Args:
        branch: name of branch to push.

    Returns:
        A `OneOf` containing an `Issue` or the response from the command.
    """
    res = subprocess.eval_cmd(f'git push -u origin "{branch}"')
    return res.flat_map_bad(hone('git push failure'))


def stash() -> Res[str]:
    """Stash the current changes in the branch.

    Returns:
        A `OneOf` containing an `Issue` or the response from `git stash`.
    """
    return subprocess.eval_cmd('git stash')


def stash_pop() -> Res[str]:
    """Pop the changes stored in the git stash.

    Returns:
        A `OneOf` containing an `Issue` or the response from `git stash pop`.
    """
    return subprocess.eval_cmd('git stash pop')


def checkout_branch(branch: str, create: bool = True) -> Res[str]:
    """Checkout a branch.

    Args:
        branch: name of branch to checkout
        create: create new branch

    Returns:
        A `OneOf` containing an `Issue` of the git response.
    """
    opt = '-b' if create else ''
    res = subprocess.eval_cmd(f'git checkout {opt} {branch}')
    return res.flat_map_bad(hone('git checkout failure'))


def get_first_commit_sha() -> Res[str]:
    """Find the first commit sha in the repository.

    Returns:
        A `OneOf` containing an `Issue` or a string of the first commit sha.
    """
    return subprocess.eval_cmd('git rev-list --max-parents=0 HEAD')


def get_current_commit_sha() -> Res[str]:
    """Find the sha of the current commit.

    Returns:
        A `OneOf` containing an `Issue` or a string of the current commit sha.
    """
    return subprocess.eval_cmd('git rev-parse HEAD')


def get_repo_path() -> Res[str]:
    """Get the absolute path to the repository.

    Returns:
        An issue or a string of the path to the repo.
    """
    return subprocess.eval_cmd('git rev-parse --show-toplevel')


def get_remote_url() -> Res[str]:
    """Find the remote url of the repo.

    Returns:
        A `OneOf` containing an `Issue` or a string with the url.
    """
    return subprocess.eval_cmd('git config --get remote.origin.url')


def get_commits(
    first: str,
    latest: str = 'HEAD',
) -> Res[list[str] | None]:
    """Get a list of all the commits between two tags.

    May return `None` in the special case when `first` is `0.0.0`.

    Args:
        first: The first tag.
        latest: The second tag, defaults to HEAD.

    Returns:
        A `OneOf` containing an `Issue` or a list of all the commits.
    """
    if first == '0.0.0':
        return Good(None)
    cmd = f'git log {latest}...{first} --oneline --no-color'
    return subprocess.eval_cmd(cmd).map(lambda out: out.splitlines())


def _extract_status(msg: str) -> tuple[str, str]:
    matches = [
        ('Untracked files', 'untracked'),
        ('Your branch is ahead', 'ahead'),
        ('Your branch is behind', 'behind'),
        ('working tree clean', 'clean'),
        ('Changes to be committed', 'staged'),
        ('Changed but not updated', 'dirty'),
        ('Changes not staged', 'dirty'),
        ('Unmerged paths', 'dirty'),
        ('diverged', 'diverged'),
    ]
    for substr, key in matches:
        if substr in msg:
            return key, substr
    return 'unknown', 'unknown'


def get_status(*, check_stash: bool = False) -> Res[tuple[str, str]]:
    """Find the current git status.

    Note that checking for stashed changes is not part of the regular git
    status. To opt in you must set `check_stash=True`.

    Args:
        check_stash: Check if there are any stashed changes.

    Returns:
        A `OneOf` containing an `Issue` or a word denoting the git status.
    """
    res = subprocess.eval_cmd('git status')
    if check_stash:
        stash_files = subprocess.eval_cmd('git stash show').get_or_else('')
        has_untracked = res.get_or_else('').startswith('Untracked files')
        if not has_untracked and stash_files:
            return Good(('stash', 'stash'))
    return one_of(
        lambda: [
            _extract_status(msg)
            for msg in res
        ],
    )


def raw_status() -> Res[str]:
    """Obtain the output of "git status".

    Returns:
        A `OneOf` containing an `Issue` or the output of "git status".
    """
    return subprocess.eval_cmd('git status')


def _list_tags(tag_response: str) -> dict[str, str]:
    lines = tag_response.strip().splitlines()
    split_lines = (line.split('\trefs/tags/') for line in lines)
    return {tag: sha for sha, tag in split_lines}


def list_tags(pattern: str) -> Res[dict[str, str]]:
    """List all the tags matching a pattern.

    Args:
        pattern: The pattern to match.

    Returns:
        A `OneOf` containing an `Issue` or a string specifying the branch.
    """
    cmd = f'git ls-remote --tags origin -l "{pattern}"'
    return one_of(
        lambda: [
            _list_tags(output)
            for output in subprocess.eval_cmd(cmd)
        ],
    )


def remove_git_tag(tag) -> Res[str]:
    """Remove a git tag.

    It is important to remove the local tag before removing the remote tag.

    Args:
        tag: The tag to remove.

    Returns:
        A `OneOf` containing an `Issue` or the response from the git command.
    """
    tag_ref = f'refs/tags/{tag}'
    return one_of(lambda: [
        f'{local}\n{remote}'
        for _ in subprocess.eval_cmd(f'git fetch origin +{tag_ref}:{tag_ref}')
        for local in subprocess.eval_cmd(f'git tag -d {tag}')
        for remote in subprocess.eval_cmd(f'git push origin :{tag_ref}')
    ])


def update_git_tag(
    tag: str,
    sha: str,
    remote_tags: list[str],
    *,
    skip: bool = False,
) -> Res[str]:
    """Create or move a git tag.

    The remote_tags should be provided to determine if we need to remove them
    before moving the tag. See https://stackoverflow.com/a/28280404 for more
    details on the git commands.

    Args:
        tag: The tag to set.
        sha: The commit sha to set the tag to.
        remote_tags: The list of remote tags.
        skip: Skip process if set to true.

    Returns:
        A `OneOf` containing an `Issue` or the response from the git command.
    """
    if skip:
        return Good(f'skipping {tag}')
    removal_output = ''
    if tag in remote_tags:
        removal_res = remove_git_tag(tag)
        if isinstance(removal_res, Bad):
            return Bad(removal_res.value)
        removal_output = f'{removal_res.value}\n'
    return one_of(lambda: [
        f'{removal_output}{local}\n{remote}'
        for local in subprocess.eval_cmd(f'git tag {tag} {sha}')
        for remote in subprocess.eval_cmd(f'git push origin {tag}')
    ])


def tag_release(version: str, sha: str, *, major_only: bool) -> Res[str]:
    """Create a git tags for a release.

    This is done to keep a major and minor versions tags pointing to the latest.

    Args:
        version: The version to tag.
        sha: The commit sha to tag.
        major_only: Only create the major tag.

    Returns:
        A `OneOf` containing an `Issue` or the response from the git command.
    """
    v_parts = version.split('.')
    # Assuming the version is valid
    major, minor = v_parts[0], v_parts[1]
    major_tag = f'v{major}'
    minor_tag = f'v{major}.{minor}'

    return one_of(lambda: [
        '\n'.join([major_out, minor_out])
        for all_tags in list_tags(f'v{major}*')
        for major_out in update_git_tag(major_tag, sha, list(all_tags))
        for minor_out in update_git_tag(
            minor_tag,
            sha,
            list(all_tags),
            skip=major_only,
        )
    ])



import re
from typing import Any, Dict, List, Optional, Tuple

from ....core import Good, Res, issue, one_of
from .io import project_stats_json, project_status_str
from .types import (
    Configuration,
    ExitCode,
    FileReport,
    ProjectStatus,
    RuleInfo,
    Transform,
    Violation,
)


def replace_filenames(
    payload: str,
    file_prefix: Optional[str],
) -> Res[str]:
    """Replace the prefix of file names.

    Args:
        payload: str
            The payload from a compiler/linter.
        file_prefix: Optional[str]
            A string of the form `[old]:[new]`. The old prefix can be
            a `|` separated list of strings. For instance
            `path/to/foo|path/to:bar`

    Returns:
        A `OneOf` containing an Issue or the `payload` with all file path
        replaced according to the file_prefix.
    """
    if not file_prefix:
        return Good(payload)
    try:
        old, prefix = file_prefix.split(':')
    except ValueError as ex:
        return issue('file_prefix param missing `:`', cause=ex)
    return Good(re.sub(
        fr'({old})(.*?)\.([a-z]+)',
        lambda x: [
            # Not sure how fix this mypy issue
            f'{prefix}{filename}.{ext}'  # type: ignore
            for _, filename, ext in (x.groups(),)
        ][0],
        payload,
    ))


def filter_reports(
    reports: List[FileReport],
    regex: Optional[str] = None,
) -> List[FileReport]:
    """Filter the list of file reports based on the file path.

    It also removes any file report that does not have any violations.

    Args:
        reports: List of file reports.
        regex: The post processor configuration.

    Returns:
        A list of file reports with at least one violation and file_path
        matching the provided regex.
    """
    return [
        rpt
        for rpt in reports
        if rpt.violations and (not regex or re.match(regex, rpt.file_path))
    ]


def to_rules_dict(reports: List[FileReport]) -> Dict[str, List[Violation]]:
    """Convert a list of `FileReport` to a map of rules to `Violation`.

    Args:
        reports: A list of `FileReport` instances.

    Returns:
        A dictionary mapping rules to a list of `Violation`.
    """
    rules: Dict[str, List[Violation]] = {}
    for report in reports:
        for violation in report.violations:
            if violation.rule_id not in rules:
                rules[violation.rule_id] = []
            rules[violation.rule_id].append(violation)
    return rules


def _process_rules_dict(
    rules_dict: Dict[str, List[Violation]],
    allowed_rules: Dict[str, int],
    ignored_rules: Dict[str, str],
    rules: Dict[str, RuleInfo],
) -> Tuple[bool, bool]:
    """Populate the `rules` dictionary.

    It is meant to be used in the `get_project_status` function.

    Args:
        rules_dict: A map of rule ids to a list of violations.
        allowed_rules: A dictionary specifying the allowed violations.
        ignored_rules: A dictionary specified the rules to ignore.
        rules: An empty dictionary created in `get_project_status`.

    Returns:
        The preliminary values for failed and needs_readjustment.
    """
    failed = False
    needs_readjustment = False

    for rule_id, violations in rules_dict.items():
        total_violations = len(violations)
        allowed = allowed_rules.get(rule_id, 0)
        ignored = bool(ignored_rules.get(rule_id))
        if not ignored:
            if total_violations > allowed:
                failed = True
            elif total_violations < allowed:
                needs_readjustment = True
        rules[rule_id] = RuleInfo(
            rule_id,
            violations,
            total_violations,
            allowed,
            ignored,
        )
    return failed, needs_readjustment


def get_project_status(
    payload: str,
    reports: List[FileReport],
    rules_dict: Dict[str, List[Violation]],
    allowed_rules: Dict[str, int],
    ignored_rules: Dict[str, str],
) -> ProjectStatus:
    """Analyze the rules_dictionary with the allowed and ignored rules.

    Args:
        payload: The original payload from the compiler/linter.
        reports: List of file reports.
        rules_dict: A map of rule ids to a list of violations.
        allowed_rules: A dictionary specifying the allowed violations.
        ignored_rules: A dictionary specifying the rules to ignore.

    Returns:
        A ProjectStatus object.
    """
    rules: Dict[str, RuleInfo] = {}
    files: Dict[str, List[Violation]] = {
        x.file_path: x.violations
        for x in reports
    }

    failed, needs_readjustment = _process_rules_dict(
        rules_dict, allowed_rules, ignored_rules, rules,
    )

    for rule_id, allowed in allowed_rules.items():
        if rule_id not in rules:
            rules[rule_id] = RuleInfo(rule_id, [], 0, allowed, ignored=False)
            if allowed > 0:
                needs_readjustment = True

    status = ExitCode.ok
    if failed:
        status = ExitCode.error
    elif needs_readjustment:
        status = ExitCode.needs_readjustment
    return ProjectStatus(status, payload, rules, files)


def process(
    raw_payload: str,
    transform: Transform,
    allowed_rules: Dict[str, int],
    ignored_rules: Dict[str, str],
    celt_config: Configuration,
) -> Res[ProjectStatus]:
    """Process the output of a compiler/linter.

    Args:
        raw_payload: The payload from the compiler/linter.
        transform: Function to generate a list of `FileReport` objects.
        allowed_rules: A dictionary specifying the allowed violations.
        ignored_rules: A dictionary specifying the rules to ignore.
        celt_config: The post processor configuration.

    Returns:
        A `OneOf` containing an `Issue` or a `ProjectStatus` object.
    """
    return one_of(lambda: [
        project_status
        for payload in replace_filenames(raw_payload, celt_config.file_prefix)
        for reports in transform(payload)
        for filtered in (filter_reports(reports, celt_config.file_regex),)
        for rules_dict in (to_rules_dict(filtered),)
        for project_status in (
            get_project_status(
                payload, filtered, rules_dict, allowed_rules, ignored_rules,
            ),
        )
    ])


class PostProcessor:
    """A post processor to handle a compiler/linter ouput."""

    def __init__(
        self,
        name: str,
        celt_config: Configuration,
        transform: Transform,
    ):
        """Instantiate a `PostProcessor`.

        Args:
            name: The name of the compiler/linter.
            celt_config: The post processor configuration.
            transform: Function to generate a list of `FileReport` objects.
        """
        self.name = name
        self.celt_config = celt_config
        self.transform = transform

    def run(
        self,
        payload: str,
        config: Dict[str, Any],
    ) -> Res[ProjectStatus]:
        """Run the processor on the given payload.

        Args:
            payload: The payload from the compiler/linter.
            config: A dictionary with rule allowance/ignores.

        Returns:
            A `OneOf` containing an `Issue` or the `ProjectStatus`.
        """
        cap_name = self.name.capitalize()
        allowed_rules = config.get(f'allowed{cap_name}Rules', {})
        if self.celt_config.ignore_error_allowance:
            allowed_rules = {}
        ignored_rules = config.get(f'ignored{cap_name}Rules', {})
        return process(
            payload,
            self.transform,
            allowed_rules,
            ignored_rules,
            self.celt_config,
        )

    def to_str(self, project: ProjectStatus) -> str:
        """Stringify a `ProjectStatus`.

        Args:
            project: The `ProjectStatus` obtained by running the `run` method.

        Returns:
            The string version of the project status.
        """
        return project_status_str(project, self.celt_config)

    def stats_json(self, project: ProjectStatus) -> str:
        """Stringify a `ProjectStatus`.

        Show a dictionary with the total number of violations. Useful when
        writing an entry for the configuration file.

        Args:
            project: The `ProjectStatus` obtained by running the `run` method.

        Returns:
            The string version of the project status.
        """
        return project_stats_json(self.name, project)



import sys
from pathlib import Path

from .one_of import Good, Issue, OneOf, issue, one_of


def read_file(filename: str | None) -> OneOf[Issue, str]:
    """FP version of open to read the contents of a file.

    If `None` is provided it will attempt to read from `sys.stdin`.

    Args:
        filename: The file path to read.

    Returns:
        A `Good` containing the contents of the file.
    """
    if filename is None:
        return Good(sys.stdin.read())
    try:
        with Path.open(Path(filename), encoding='UTF-8') as fp:
            return Good(fp.read())
    except Exception as ex:
        return issue(
            'failed to read file',
            context={'filename': filename},
            cause=ex,
        )


def write_file(filename: str, text: str) -> OneOf[Issue, int]:
    """FP version of open to write to a file.

    Args:
        filename: The file path where contents will be written.
        text: The contents of the file.

    Returns:
        A `Good` containing 0 if the file was written.
    """
    try:
        with Path.open(Path(filename), 'w', encoding='UTF-8') as fp:
            fp.write(text)
    except Exception as ex:
        return issue(
            'failed to write file',
            context={'filename': filename},
            cause=ex,
        )
    return Good(0)


def assert_file_exists(path: str) -> OneOf[Issue, Path]:
    """Assert that a file exists.

    Args:
        path: The path to the file.

    Returns:
        A `Good` containing the `Path` if the file exists.
    """
    path_inst = Path(path)
    if path_inst.exists():
        return Good(path_inst)
    return issue('file does not exist', context={'path': path})


def _insert_to_file(
    file_content: str,
    start: str,
    text: str,
    end: str,
) -> OneOf[Issue, str]:
    first_split = file_content.split(start)
    left_content = first_split[0]
    right_content = ''
    if len(first_split) > 1:
        second_split = first_split[-1].split(end)
        right_content = second_split[-1]
    file_data = [left_content, start, text, end, right_content]
    return Good(''.join(file_data))


def insert_to_file(
    filename: str,
    start: str,
    text: str,
    end: str,
) -> OneOf[Issue, int]:
    """Insert content to a file.

    Args:
        filename: The file to insert to.
        start: The start delimiter where the insertion will take place.
        text: The main content to insert.
        end: The end delimiter.

    Returns:
        None if successful, else an issue.
    """
    return one_of(lambda: [
        0
        for file_content in read_file(filename)
        for new_content in _insert_to_file(file_content, start, text, end)
        for _ in write_file(filename, new_content)
    ])



from m.cli import Meta, cli_integration_token

meta = Meta(
    help='call the github api',
    description="""
        The following commands make calls to the Github api.
    """,
)
add_arguments = cli_integration_token('github', 'GITHUB_TOKEN')



import logging
import sys


class JsonFileHandler(logging.FileHandler):
    """A log handler to output to a file."""

    def __init__(self, formatter: logging.Formatter, filename: str):
        """Initialize the handler.

        Args:
            formatter: The formatter to use with the handler.
            filename: The file where to store the log recrods.
        """
        super().__init__(filename, encoding='UTF-8')
        self.setFormatter(formatter)

    def filter(self, _record: logging.LogRecord) -> bool:
        """Display all records.

        Args:
            _record: A logging record instance.

        Returns:
            True
        """
        return True


class StdErrHandler(logging.StreamHandler):
    """A log handler to print to stderr."""

    def __init__(self, formatter: logging.Formatter):
        """Initialize the handler.

        Args:
            formatter: The formatter to use with the handler.
        """
        super().__init__(sys.stderr)
        self.setFormatter(formatter)

    def filter(self, record: logging.LogRecord) -> bool:
        """Only display warnings or error records.

        Args:
            record: A logging record instance.

        Returns:
            True if the record is a warning or error.
        """
        return record.levelno >= logging.WARNING


class StdOutHandler(logging.StreamHandler):
    """A log handler to print to stdout."""

    def __init__(self, formatter: logging.Formatter):
        """Initialize the handler.

        Args:
            formatter: The formatter to use with the handler.
        """
        super().__init__(sys.stdout)
        self.setFormatter(formatter)

    def filter(self, record: logging.LogRecord) -> bool:
        """Display non-warnings and non-error records.

        Args:
            record: A logging record instance.

        Returns:
            True if the record is not a warning or error.
        """
        return record.levelno < logging.WARNING



import inspect
import json
import sys
import traceback
from collections import OrderedDict
from contextlib import suppress
from typing import cast

from m.color import highlight_json, highlight_yaml
from typing_extensions import TypedDict

from . import yaml

IssueDict = TypedDict(
    'IssueDict',
    {
        'message': str,
        'description': str,
        'cause': object,
        'context': object,
        'traceback': list[str] | str,
    },
    total=False,
)


def remove_traceback(issue_dict: object) -> None:
    """Remove the `traceback` key from a dictionary if it exists.

    It will recursively remove the `traceback` from its cause or context.

    Args:
        issue_dict: A dictionary representation of an `Issue`.
    """
    if isinstance(issue_dict, dict):
        if 'traceback' in issue_dict:
            issue_dict.pop('traceback')
        cause = issue_dict.get('cause')
        if cause:
            remove_traceback(cause)
        context = issue_dict.get('context')
        remove_traceback(context)
        if isinstance(context, list):
            for context_item in context:
                remove_traceback(context_item)


def _traceback_to_str(traceback_list: list[str]) -> str:
    if not traceback_list:
        return ''
    return '\n'.join([
        '(most recent call last):',
        *traceback_list,
        '',
    ])


# Rule WPS230 is suppressed here since I want to have a flat structure
# with this class. In other places we can refactor and group attributes into
# other classes.
class Issue(Exception):  # noqa: N818, WPS230 - Intention is not to raise
    """Wrapper to keep track of all exceptions.

    It provides a 'cause' field so that we may know why an issue was
    triggered.
    """

    show_traceback = True
    yaml_traceback = True

    message: str
    description: str | None
    cause: Exception | None
    context: object | None
    include_traceback: bool
    cause_tb: list[str] | None
    traceback: list[str] | None

    def __init__(  # noqa: WPS211 - need to initialize the attributes
        self,
        message: str,
        description: str | None = None,
        cause: Exception | None = None,
        context: object | None = None,
        include_traceback: bool = True,
    ):
        """Create an Issue.

        Args:
            message: Simple description of the issue.
            description: More in depth detail on the issue
            cause: The exception/Issue that is responsible for this instance.
            context: Dictionary with any useful data related to the issue.
            include_traceback: If False, it skips computing the traceback.
        """
        super().__init__()
        self.message = message
        self.description = description
        self.cause = cause
        if cause and not isinstance(cause, Issue):
            # https://stackoverflow.com/a/12539332/788553
            with suppress(BaseException):
                exc_info = sys.exc_info()
                fmt_exception = (
                    traceback.format_exception_only(exc_info[0], exc_info[1])
                    if exc_info[0] is not None and exc_info[1] is not None
                    else []
                )
                exception_list = [
                    *traceback.format_tb(exc_info[2]),
                    *fmt_exception,
                ]
                self.cause_tb = [
                    y
                    for x in exception_list
                    for y in x.splitlines()
                ]
        self.context = context
        self.include_traceback = include_traceback
        self.traceback = None
        if self.include_traceback:
            frame = inspect.currentframe()
            self.traceback = [
                y
                for x in traceback.format_stack(frame)[:-1]
                for y in x.splitlines()
            ]

    def only_context(self) -> bool:
        """Return true if the issue only offers context.

        In some cases we may create an issue with only a message and a
        context. This function will let us know of such case so that a
        log formatter may be able to unwrap the context.

        Returns:
            True if it is safe to only display the context value on a log.
        """
        has_context = self.context is not None
        all_props = (self.description, self.traceback, self.cause)
        return has_context and not [_ for _ in all_props if _]

    def to_dict(self, show_traceback: bool = True) -> IssueDict:
        """Convert to a ordered dictionary.

        This is done so that each of the properties are written in an expected
        order.

        Args:
            show_traceback: If False, it will remove all stacktraces.

        Returns:
            An `IssueDict` instance.
        """
        issue_dict = cast(IssueDict, OrderedDict())
        issue_dict['message'] = self.message
        if self.description:
            issue_dict['description'] = self.description
        if self.context:
            issue_dict['context'] = self.context
        if self.include_traceback and self.traceback:
            issue_dict['traceback'] = self.traceback
            if Issue.yaml_traceback and isinstance(self.traceback, list):
                issue_dict['traceback'] = _traceback_to_str(self.traceback)
        if self.cause:
            self._handle_cause(issue_dict)
        if not show_traceback:
            remove_traceback(issue_dict)
        return issue_dict

    def to_str(self, show_traceback: bool) -> str:
        """Convert the instance to string.

        Args:
            show_traceback: If false, it will remove the error traceback.

        Returns:
            A string representation of the Issue instance.
        """
        issue_dict = self.to_dict(show_traceback=show_traceback)
        if Issue.yaml_traceback:
            return yaml.dumps(issue_dict)
        return json.dumps(issue_dict, indent=2)

    def __str__(self) -> str:
        """Convert the instance to a string.

        Returns:
            A string representation of the Issue instance.
        """
        issue_str = self.to_str(Issue.show_traceback)
        if Issue.yaml_traceback:
            return highlight_yaml(issue_str)
        return highlight_json(issue_str)

    def _handle_cause(self, issue_dict: IssueDict) -> None:
        if isinstance(self.cause, Issue):
            issue_dict['cause'] = self.cause.to_dict()
        else:
            issue_dict['cause'] = {
                'message': str(self.cause),
                'traceback': self.cause_tb,
            }
            # I'm clearly assigning to a dict in the above statement...
            # but it is an object so a dict can be assigned to it but that
            # does not mean that we can use `pop` on an object. So we have to
            # help mypy know that know we are dealing with a dict.
            cause = cast(dict, issue_dict['cause'])
            if not self.cause_tb:
                cause.pop('traceback', None)
            if Issue.yaml_traceback and isinstance(self.cause_tb, list):
                cause['traceback'] = _traceback_to_str(self.cause_tb)



from m.cli import Arg, BaseModel, command


class Arguments(BaseModel):
    """Close and immediately open another block."""

    to_close: str = Arg(
        help='block name to close',
        positional=True,
        required=True,
    )
    name: str = Arg(
        help='block name to open',
        positional=True,
        required=True,
    )
    description: str = Arg(
        help='block description',
        positional=True,
        required=True,
    )


@command(
    help='close and open a sibling block',
    model=Arguments,
)
def run(arg: Arguments):
    from m.log import Logger

    logger = Logger('m.cli')
    logger.close_block(arg.to_close)
    logger.open_block(arg.name, arg.description)
    return 0



from m.cli import Arg, BaseModel, command, run_main
from m.cli.handlers import create_dict_handler
from m.core.io import env


class Arguments(BaseModel):
    """Retrieve pull requests associated with a branch.

    example::

        $ m github branch_prs --owner jmlopez-rod --repo m release/0.18.0
    """

    pretty: bool = Arg(
        default=False,
        help='format json payload with indentation',
    )
    yaml: bool = Arg(
        default=False,
        help='use yaml format',
    )
    owner: str = Arg(
        default=env('GITHUB_REPOSITORY_OWNER'),
        help='repo owner',
    )
    repo: str = Arg(
        help='repo name',
        required=True,
    )
    branch: str = Arg(
        help='branch name',
        required=True,
        positional=True,
    )


@command(
    help='get prs associated with a branch',
    model=Arguments,
)
def run(arg: Arguments, arg_ns) -> int:
    from m.github.graphql.queries.branch_prs import fetch_raw

    return run_main(
        lambda: fetch_raw(
            arg_ns.token,
            arg.owner,
            arg.repo,
            arg.branch,
        ),
        result_handler=create_dict_handler(arg.pretty, as_yaml=arg.yaml),
    )



import re


def docker_tags(m_tag: str, *, skip_floating: bool = False) -> list[str]:
    """Convert an m_tag to docker tags.

    Floating tags include `latest` and any `vX` or `vX.Y` tags.
    The `skip_floating` argument is intended for use when building packages
    that have already been published in the past.

    Args:
        m_tag: A tag/version provided by m.
        skip_floating: If true, do not include floating tags.

    Returns:
        A tag that can be used by docker during the publishing step.
    """
    tags = []
    regex = r'\d+.\d+.\d+-(.*)\.(.*)'
    matches = re.match(regex, m_tag)
    if matches:
        tag, _ = matches.groups()
        if tag.startswith('rc') or tag.startswith('hotfix'):
            tags.append('next')
            index = 2 if tag.startswith('rc') else 6
            pr_num = tag[index:]
            tags.append(f'pr{pr_num}')
        else:
            tags.append(tag)
    else:
        x, y, _ = m_tag.split('.')
        if not skip_floating:
            tags.append('latest')
            tags.append(f'v{x}')
            tags.append(f'v{x}.{y}')

    return tags



from typing import Any

from m.cli import (
    Arg,
    command,
    create_issue_handler,
    run_main,
    validate_json_payload,
)
from pydantic import BaseModel


class Arguments(BaseModel):
    r"""Query json data.

    single value::

        m jsonq path.to.property < file.json
        cat file.json | m jsonq path.to.property
        m jsonq @file.json path.to.property

    Return the value stored in the json file. For arrays and objects it
    will print the python representation of the object.

    multiple values::

        m jsonq path1 path2 path3 < file.json
        cat file.json | m jsonq path1 path2 path3
        m jsonq @file.json path1 path2 path3

    use `read` to store in bash variables::

        read -r -d '\n' \
            var1 var2 var3 \
            <<< "$(m jsonq @file.json 'path1' 'path2' 'path3')"

    If available, take advantage of better queries through jq and yq:

    - jq: https://stedolan.github.io/jq/manual/
    - yq: https://mikefarah.gitbook.io/yq/
    """

    payload: Any = Arg(
        default='@-',
        help='json data: @- (stdin), @filename (file), string',
        validator=validate_json_payload,
        positional=True,
    )

    query: list[str] = Arg(
        help='path to json data',
        positional=True,
        nargs='+',
    )

    warn: bool = Arg(
        default=False,
        help='print warning messages instead of errors',
        aliases=['w', 'warn'],
    )

    separator: str = Arg(
        default='\n',
        help='separator for multiple values',
        aliases=['s', 'separator'],
    )


@command(
    help='query json data',
    model=Arguments,
)
def run(arg: Arguments):
    from m.core.json import jsonq

    return run_main(
        lambda: jsonq(arg.payload, arg.separator, *arg.query),
        result_handler=print,
        issue_handler=create_issue_handler(arg.warn),
    )



from ..misc import argument_description
from ..types import MISSING, AnyMap, FuncArgs


def should_handle(field: AnyMap) -> bool:
    """Handle a positional field.

    Args:
        field: A dictionary with information for a cli argument.

    Returns:
        True if it should handle the field as positional
    """
    return field.get('positional', False) is True


def handle_field(name: str, field: AnyMap) -> FuncArgs:
    """Treat the field as a positional field.

    Args:
        name: The name of the argument.
        field: A dictionary representation of the field.

    Returns:
        Function arguments for the parser `add_argument` method.
    """
    default = field.get('default', MISSING)
    validator = field.get('validator', None)
    is_required = field.get('required', False)
    nargs = field.get('nargs', None)

    arg_default = default if default is MISSING else repr(default)
    args: AnyMap = {
        'help': argument_description(field['description'], arg_default),
    }
    if default is not MISSING:
        args['default'] = default

    if validator:
        args['type'] = validator

    if nargs:
        args['nargs'] = nargs
    elif not is_required:
        args['nargs'] = '?'

    return FuncArgs(args=[name], kwargs=args)



import os

from m.core import Bad, Good, OneOf
from pydantic import BaseModel


class MissingEnvVars(BaseModel):
    """Container for missing environment variables."""

    # list of environment variables that are required but do not exist
    missing: list[str]

    # environment variables that are defined
    defined: dict[str, str]


def require_env_vars(env_vars: list[str]) -> OneOf[MissingEnvVars, None]:
    """Require that the given environment variables are defined.

    Args:
        env_vars: list of environment variables to check.

    Returns:
        None if all environment variables are defined, otherwise a
        MissingEnvVars instance.
    """
    missing: list[str] = []
    defined: dict[str, str] = {}
    for env_var in env_vars:
        env_var_value = os.environ.get(env_var)
        if env_var_value is None:
            missing.append(env_var)
        else:
            defined[env_var] = env_var_value
    if missing:
        return Bad(MissingEnvVars(missing=missing, defined=defined))
    return Good(None)


class DevcontainerEnvVars(BaseModel):
    """Collection of variables required for devcontainer operations.

    All properties are prefixed with `MDC` to denote that those are variables
    generated by `M Dev Container`.
    """

    mdc_repo: str
    mdc_workspace: str
    mdc_pnpm_workspace: str
    mdc_venv_workspace: str

    def to_bash(self: 'DevcontainerEnvVars') -> str:
        """Create a bash code snippet that can be used in a bashrc file.

        Returns:
            A `.bashrc` code snippet.
        """
        dict_data = self.model_dump()
        return '\n'.join([
            f"export {var_name.upper()}='{var_value}'"
            for var_name, var_value in dict_data.items()
        ])


def devcontainer_env_vars() -> DevcontainerEnvVars:
    """Get the environment variables for a devcontainer.

    This function is intended to be used in a devcontainer or in a Github
    runner. It will return the following values::

        - repo: The repository name.
        - workspace: The path to directory containing the repo files.
        - pnpmWorkspace: The path where pnpm operations should take place.
        - venvWorkspace: The path where venv operations should take place.

    Returns:
        A dictionary with the values described above.
    """
    github = os.environ.get('GITHUB_WORKSPACE')
    container = os.environ.get('CONTAINER_WORKSPACE')
    workspace = github or container or 'ERROR_UNKNOWN_WORKSPACE'
    # big assumption here that the name of the repo is reflected in the name
    # of the containing directory. Trying to avoid having to call `git`.
    repo_name = workspace.split('/')[-1]
    pnpm_workspace = f'/opt/pnpm/{repo_name}'
    venv_workspace = f'/opt/venv/{repo_name}'
    return DevcontainerEnvVars(
        mdc_repo=repo_name,
        mdc_workspace=workspace,
        mdc_pnpm_workspace=pnpm_workspace,
        mdc_venv_workspace=venv_workspace,
    )




from pygments import highlight
from pygments.formatters.terminal256 import Terminal256Formatter
from pygments.lexer import Lexer
from pygments.lexers.data import YamlLexer
from pygments.lexers.web import JsonLexer
from pygments.styles import get_style_by_name

from .disable import color_disabled

style = get_style_by_name('github-dark')
terminal_formatter = Terminal256Formatter(style=style)


def _highlight(text: str, lexer: Lexer) -> str:
    if color_disabled():
        return text
    try:
        return highlight(text, lexer=lexer, formatter=terminal_formatter)
    except Exception:
        return text


def highlight_json(text: str) -> str:
    """Highlight a json string.

    Args:
        text: The json string to highlight.

    Returns:
        A colorized json string.
    """
    return _highlight(text, JsonLexer())


def highlight_yaml(text: str) -> str:
    """Highlight a yaml string.

    Args:
        text: The yaml string to highlight.

    Returns:
        A colorized yaml string.
    """
    return _highlight(text, YamlLexer())



from m.cli import ArgProxy, BaseModel, command


class Arguments(BaseModel):
    """Print a bash snippet of setup instructions.

    This snippet can be used in a `.bashrc` file to setup the environment.
    It will also provide aliases that call other `m` commands to help
    facilitate the development in a devcontainer workflow.
    """

    section: str | None = ArgProxy(
        '--section',
        choices=['env', 'devex', 'venv'],
        help='section to print out',
    )


@command(
    help='print out a bash snippet that exports variables',
    model=Arguments,
)
def run(args: Arguments) -> int:
    from m.devcontainer.bashrc import (
        bashrc_snippet,
        devex_snippet,
        venv_snippet,
    )
    from m.devcontainer.env import devcontainer_env_vars

    if args.section:
        sections = {
            'env': devcontainer_env_vars().to_bash(),
            'devex': devex_snippet,
            'venv': venv_snippet,
        }
        print(sections.get(args.section, ''))  # noqa: WPS421
        return 0

    snippet = [
        devcontainer_env_vars().to_bash(),
        bashrc_snippet,
    ]
    print('\n'.join(snippet))  # noqa: WPS421
    return 0



from m.cli import ArgProxy, BaseModel, command, run_main
from m.core.io import env


class Arguments(BaseModel):
    """End the release process.

    Merges the release/hotfix branch into the master branch.
    """

    github_token: str = ArgProxy(
        '--github-token',
        default=env('GITHUB_TOKEN'),
        help='Github PAT (default: env.GITHUB_TOKEN)',
    )


@command(
    help='end the release process',
    model=Arguments,
)
def run(arg: Arguments):
    from m.ci.end_release import end_release

    return run_main(lambda: end_release(arg.github_token))



from inspect import cleandoc
from typing import Any, Callable, Literal

from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined

from .engine.types import FuncArgs

_Unset: Any = PydanticUndefined


def RemainderArgs(  # noqa: N802
    *,
    help: str | None = _Unset,  # noqa: WPS125
) -> Any:
    """Provide a list of unrecognized arguments.

    This is a escape hatch and does not provide any typings. May be
    useful for commands that need to pass arguments to other commands.

    Args:
        help: Human-readable description.

    Returns:
        A new [`FieldInfo`][pydantic.fields.FieldInfo], the return annotation is
        `Any` so `Arg` can be used on type annotated fields without causing a typing error.
    """
    return FieldInfo.from_field(
        [],
        description=help,
        json_schema_extra={'__remainder_args': True},
    )


def Arg(  # noqa: N802, WPS211
    default: Any = PydanticUndefined,
    *,
    help: str | None = _Unset,  # noqa: WPS125
    positional: bool | None = None,
    required: bool | None = None,
    aliases: list[str] | None = None,
    nargs: int | Literal['?', '*', '+'] | None = None,
    validator: Callable[[str], str] | None = None,
) -> Any:
    """Create a pydantic `Field`.

    Field docs: https://docs.pydantic.dev/2.2/usage/fields

    Defines properties used to create an argparse argument. This function
    should work for most cases. If we need something that is not covered
    we can use `ArgProxy` instead which is untyped but provides all the arguments
    and keyword arguments to argparse.

    See https://docs.python.org/3/library/argparse.html.

    Args:
        default: Default value if the field is not set.
        help: Human-readable description.
        positional: Whether the argument is positional or not.
        required: Indicate whether an argument is required or optional.
        aliases: Alternative names for the argument.
        nargs: Number of times the argument can be used.
        validator: Function to validate the argument.

    Returns:
        A new [`FieldInfo`][pydantic.fields.FieldInfo], the return annotation is
        `Any` so `Arg` can be used on type annotated fields without causing a typing error.
    """
    extras = {
        'positional': positional,
        'validator': validator,
        'aliases': aliases,
        'nargs': nargs,
        'required': required,
    }
    return FieldInfo.from_field(
        default,
        description=help,
        json_schema_extra={k: v for k, v in extras.items() if v is not None},
    )


def ArgProxy(*args, **kwargs) -> Any:  # noqa: N802
    """Wrap function to provide all argparse inputs.

    This is a escape hatch and does not provide any typings.
    See https://docs.python.org/3/library/argparse.html.

    Args:
        args: The arguments to argparse add_arguments.
        kwargs: The keyword arguments to argparse add arguments.

    Returns:
        A new [`FieldInfo`][pydantic.fields.FieldInfo], the return annotation is
        `Any` so `Arg` can be used on type annotated fields without causing a typing error.
    """
    return FieldInfo.from_field(
        json_schema_extra={
            'proxy': FuncArgs(args=list(args), kwargs=kwargs),
        },
    )


def Meta(  # noqa: N802
    *,
    help: str,  # noqa: WPS125
    description: str,
) -> dict[str, str]:
    """Create the meta dictionary for a subcommand description.

    In the case of the root command the `help` may be set to empty since it
    is not used.

    Args:
        help: The help message for the subcommand.
        description: The description for the command/subcommand.

    Returns:
        A dictionary with the help and description.
    """
    return {
        'help': help,
        'description': cleandoc(description),
    }




from typing import Generic, TypeVar

from pydantic import BaseModel

G_Item = TypeVar('G_Item')


def identity(x: G_Item) -> G_Item:
    """Identity function.

    Args:
        x: The input

    Returns:
        The input value.
    """
    return x


class WithNodes(BaseModel, Generic[G_Item]):
    """An object that may contain nodes."""

    nodes: list[G_Item]



from functools import partial
from typing import Any, Callable

from m.core import Issue, OneOf, one_of
from m.core.json import get
from m.pydantic import CamelModel
from pydantic import BaseModel, TypeAdapter

from ..api import graphql
from ..enums import MergeableState, PullRequestReviewState
from ..generics import G_Item, WithNodes, identity


class Actor(BaseModel):
    """An object representing a committer."""

    login: str


class PullRequestReview(BaseModel):
    """A review object for a given pull request."""

    author: Actor
    body: str
    state: PullRequestReviewState


class PullRequest(CamelModel):
    """A repository pull request."""

    closed: bool
    title: str
    number: int
    base_ref_name: str
    mergeable: MergeableState
    merged: bool
    author: Actor
    latest_reviews: WithNodes[PullRequestReview]
    url: str


def _fetch(
    transform: Callable[[Any], G_Item],
    token: str,
    owner: str,
    repo: str,
    branch: str,
) -> OneOf[Issue, G_Item]:
    """Retrieve the pr information for the specified branch.

    Args:
        transform: Function to apply to the graphql result.
        token: A Github PAT.
        owner: The owner of the repo.
        repo: The name of the repo.
        branch: Name of the branch.

    Returns:
        A list of pull requests connected to the branch or an issue.
    """
    query = """query ($owner: String!, $repo: String!, $branch: String!) {
        repository(owner:$owner, name:$repo) {
            pullRequests(first: 2, headRefName: $branch) {
                nodes {
                    closed
                    title
                    number
                    baseRefName
                    mergeable
                    merged
                    author {
                        login
                    }
                    latestReviews(last: 10) {
                        nodes {
                            author {
                                login
                            }
                            body
                            state
                        }
                    }
                    url
                }
            }
        }
    }"""
    variables = {'owner': owner, 'repo': repo, 'branch': branch}
    return one_of(
        lambda: [
            transform(pull_requests)
            for res in graphql(token, query, variables)
            for pull_requests in get(res, 'repository.pullRequests.nodes')
        ],
    )


fetch_raw = partial(_fetch, identity)


def fetch(
    token: str,
    owner: str,
    repo: str,
    branch: str,
) -> OneOf[Issue, list[PullRequest]]:
    """Retrieve the pr information for the specified branch.

    Args:
        token: A Github PAT.
        owner: The owner of the repo.
        repo: The name of the repo.
        branch: Name of the branch.

    Returns:
        A list of pull requests connected to the branch or an issue.
    """
    return _fetch(
        lambda raw: TypeAdapter(list[PullRequest]).validate_python(raw),
        token,
        owner,
        repo,
        branch,
    )



from m.cli import Arg, command, run_main
from pydantic import BaseModel


class Arguments(BaseModel):
    """Setup pnpm in the devcontainer.

    This command is meant run from inside the devcontainer. It will
    create several symbolic links so that the pnpm data may be stored in a
    volume and thus be able to be shared with other devcontainers.
    """

    work_dir: str = Arg(
        help='the work directory containing package.json',
        positional=True,
        required=True,
    )

    pnpm_dir: str = Arg(
        help='the directory where pnpm data will be stored',
        positional=True,
        required=True,
    )


@command(
    help='setup pnpm',
    model=Arguments,
)
def run(arg: Arguments):
    from m.devcontainer.pnpm import pnpm_setup

    return run_main(
        lambda: pnpm_setup(arg.work_dir, arg.pnpm_dir),
    )



from typing import Any

from ..core import one_of
from ..core.fp import Good, OneOf
from ..core.issue import Issue
from ..core.json import get
from .ci import create_ci_query
from .graphql.api import graphql


def get_pr_info(
    token: str,
    owner: str,
    repo: str,
    pr_number: int,
    file_count: int,
) -> OneOf[Issue, Any]:
    """Retrieve the information of the given Github PR.

    Args:
        token: A Github PAT.
        owner: The owner of the repo.
        repo: The name of the repo.
        pr_number: The pull request number in question.
        file_count: The maximum number of files in the pr to retrieve.

    Returns:
        The pull request information.
    """
    query = create_ci_query(
        pr_number,
        include_commit=False,
        include_release=False,
    )
    variables = {
        'owner': owner,
        'repo': repo,
        'pr': pr_number,
        'fc': file_count,
    }
    return one_of(
        lambda: [
            pr_info
            for res in graphql(token, query, variables)
            for pr_info in get(res, 'repository.pullRequest')
        ],
    )


def get_latest_release(
    token: str,
    owner: str,
    repo: str,
) -> OneOf[Issue, str]:
    """Retrieve the latest release for a repo.

    Args:
        token: A Github PAT.
        owner: The owner of the repo.
        repo: The name of the repo.

    Returns:
        The latest release.
    """
    query = """query ($owner: String!, $repo: String!) {
      repository(owner:$owner, name:$repo) {
         releases(last: 1, orderBy: {field: CREATED_AT, direction: ASC}) {
            nodes {
                name
                tagName
                publishedAt
            }
        }
      }
    }"""
    variables = {'owner': owner, 'repo': repo}
    return one_of(
        lambda: [
            tag_name
            for res in graphql(token, query, variables)
            for releases in get(res, 'repository.releases.nodes')
            for tag_name in _get_latest_release(releases)
        ],
    )


def _get_latest_release(releases: list[dict[str, str]]) -> OneOf:
    if releases:
        return get(releases[0], 'tagName')
    return Good('0.0.0')



from m.cli import BaseModel, command, run_main


class Arguments(BaseModel):
    """Create the necessary files for an m project."""


@command(
    help='initialize an m project',
    model=Arguments,
)
def run():
    from m.ci.init import init_repo

    return run_main(init_repo)



from m.cli import BaseModel, command, run_main


class Arguments(BaseModel):
    """Display the current git branch name.

    example::

        $ m git branch
        master
    """


@command(
    help='display the current git branch',
    model=Arguments,
)
def run():
    from m import git
    return run_main(git.get_branch, print)



import re
from pathlib import Path
from textwrap import dedent
from typing import List, Tuple

from m.core import Good, Res, issue, one_of
from m.core import rw as mio
from m.core import subprocess
from m.git import get_remote_url
from m.log import Logger

logger = Logger('m.cli.init')


def parse_ssh_url(ssh_url: str) -> Res[Tuple[str, str]]:
    """Find the owner and repo from an ssh url.

    Args:
        ssh_url: The url of the repo

    Returns:
        A tuple with the owner and repo (or an Issue).
    """
    match = re.findall('.*:(.*)/(.*).git', ssh_url)
    if match and match[0] and match[0][0] and match[0][1]:
        return Good(match[0])
    return issue(
        'unable to obtain owner and repo',
        context={'ssh_url': ssh_url},
    )


def get_repo_info() -> Res[Tuple[str, str]]:
    """Get the owner and repo name from the current repository.

    Returns:
        A tuple with the owner and repo (or an Issue).
    """
    return one_of(lambda: [
        owner_repo
        for ssh_url in get_remote_url()
        for owner_repo in parse_ssh_url(ssh_url)
    ])


def m_config_body(owner: str, repo: str) -> str:
    """Create the basic contents of a m configuration file.

    Args:
        owner: The repo owner.
        repo: The repo name.

    Returns:
        A yaml string to be the content of the `m.yaml` file.
    """
    body = f"""\
        owner: {owner}
        repo: {repo}
        version: 0.0.0
        workflow: m_flow
    """
    return dedent(body)


def create_m_config() -> Res[str | None]:
    """Create the m configuration file.

    Returns:
        The name of the m file or none if successful. Otherwise an `Issue`.
    """
    file_name = 'm/m.yaml'
    if Path.exists(Path(file_name)):
        logger.warning(f'{file_name} already exists')
        return Good(None)
    m_dir = Path('m')
    if not Path.exists(m_dir):
        Path.mkdir(m_dir, parents=True)
    return one_of(lambda: [
        file_name
        for owner, repo in get_repo_info()
        for _ in mio.write_file(file_name, m_config_body(owner, repo))
        for _ in logger.info(f'created {file_name}', {
            'owner': owner,
            'repo': repo,
        })
    ])


def create_changelog() -> Res[str | None]:
    """Create the changelog file.

    Returns:
        A `OneOf` containing 0 if successful or an `Issue`.
    """
    body = """\
        # Changelog

        The format of this changelog is based on
        [Keep a Changelog](http://keepachangelog.com/en/1.0.0/). The project adheres to
        [Semantic Versioning](http://semver.org/spec/v2.0.0.html)

        > Major version zero (0.y.z) is for initial development. Anything may change at
        > any time. The public API should not be considered stable.

        ## [Unreleased]
    """
    file_name = 'CHANGELOG.md'
    if Path.exists(Path(file_name)):
        logger.warning(f'{file_name} already exists')
        return Good(None)
    return one_of(lambda: [
        file_name
        for _ in mio.write_file('CHANGELOG.md', dedent(body))
        for _ in logger.info(f'created {file_name}')
    ])


def _update_gitignore(file_name: str, body: str) -> Res[str | None]:
    """List of things that should be in a gitignore file.

    Args:
        file_name: The name of the gitignore file.
        body: The current contents of the gitignore file.

    Returns:
        The contents of the new gitignore file.
    """
    if 'm/.m' in body:
        logger.warning(f'{file_name} already ignores m/.m')
        return Good(None)
    buffer: List[str] = body.splitlines()
    buffer.append('m/.m')
    entries = '\n'.join(buffer)
    return one_of(lambda: [
        file_name
        for _ in mio.write_file(file_name, f'{entries}\n')
        for _ in logger.info(f'updated {file_name}')
    ])


def update_gitignore() -> Res[str | None]:
    """Update the gitignore file.

    Adds the m/.m directory to the list.

    Returns:
        A `OneOf` containing file name or an Issue if it was unable to update
        the file.
    """
    file_name = '.gitignore'
    return one_of(lambda: [
        fname
        for _ in subprocess.eval_cmd(f'touch {file_name}')
        for body in mio.read_file(file_name)
        for fname in _update_gitignore(file_name, body)
    ])


def init_repo() -> Res[None]:
    """Initialize a repository with the basic project configurations.

    Returns:
        A `OneOf` containing 0 if successful or an `Issue`.
    """
    return one_of(lambda: [
        None
        for m_file in create_m_config()
        for gitignore in update_gitignore()
        for changelog in create_changelog()
        for _ in logger.info('setup complete', {
            'modified_files': [x for x in (m_file, gitignore, changelog) if x],
        })
    ])



from m.cli import BaseModel, command, run_main


class Arguments(BaseModel):
    """Display the current commit sha.

    example::

        $ m git current_sha
        74075a3ea5c9252a0f2b9fd6b095567b3b9b4028
    """


@command(
    help='display the current commit sha',
    model=Arguments,
)
def run():
    from m import git
    return run_main(git.get_current_commit_sha, print)



import logging
from dataclasses import dataclass
from typing import Callable

from m.core import Issue, OneOf
from pydantic import BaseModel, ConfigDict, Field


class EnvVars(BaseModel):
    """Class to store the values of the environment variables."""

    # pylint: disable=too-many-instance-attributes
    ci_env: bool = False
    github_token: str = ''
    server_url: str = ''
    run_id: str = ''
    run_number: str = ''
    run_url: str = ''
    git_branch: str = ''
    git_sha: str = ''
    triggered_by: str = ''
    triggered_by_email: str = ''
    triggered_by_user: str = ''


class Message(BaseModel):
    """Parameters needed to deliver a warning or error message."""

    msg: str = Field(description='message to display')
    title: str | None = Field(default=None, description='custom title')
    file: str | None = Field(  # noqa: WPS110 - required by Github
        default=None,
        description='filename',
    )
    line: str | None = Field(
        default=None,
        description='line number, starting at 1',
    )
    end_line: str | None = Field(default=None, description='end line number')
    col: str | None = Field(
        default=None,
        description='column number, starting at 1',
    )
    end_col: str | None = Field(default=None, description='end column number')

    model_config = ConfigDict(arbitrary_types_allowed=True)


@dataclass
class ProviderModule:
    """Container to store functions from the providers."""

    ci: bool
    env_vars: Callable[[], OneOf[Issue, EnvVars]]
    formatter: Callable[
        [logging.Formatter, logging.LogRecord, bool, bool],
        str,
    ]



import platform

from m.core import Bad, Good, Res
from m.core.rw import read_file
from pydantic import BaseModel

from .docker_build import DockerBuild
from .env import MEnvDocker

BASH_SHEBANG = '#!/bin/bash'
SET_STRICT_BASH = 'set -euxo pipefail'
AMD64 = 'amd64'
ARM64 = 'arm64'


def get_arch() -> str:
    """Get the architecture of the current machine.

    Returns:
        The architecture of the current machine.
    """
    mapping = {
        # amd
        'i386': AMD64,
        'x86_64': AMD64,
        # arm
        'aarch64': ARM64,
    }
    arch = platform.machine()
    return mapping.get(arch, arch)


class DockerImage(BaseModel):
    """Information describing how to a docker image."""

    # String used to describe the step in the workflow.
    step_name: str

    # The name of the image - usually prefixed with the name of the repo.
    image_name: str

    # Name of the docker file to use for the build step.
    docker_file: str

    # Name of a target stage to build. Leave empty to build the whole file.
    target_stage: str | None = None

    # Arguments to pass to the docker build command, they will only be injected
    # if they appear within the docker file.
    build_args: dict[str, str] = {}

    # name of envvars to be treated as secrets.
    env_secrets: list[str] | None = None

    def format_build_args(
        self: 'DockerImage',
        m_env: MEnvDocker,
        arch: str,
        extras: dict[str, str] | None = None,
    ) -> Res[list[str]]:
        """Format the arguments to pass to the docker build command.

        Docker will only inject build args if they appear in the docker file.

        Args:
            m_env: The MEnvDocker instance with the environment variables.
            arch: The architecture to build for.
            extras: Extra build arguments.

        Returns:
            A list with the build arguments.
        """
        docker_file = f'{m_env.base_path}/{self.docker_file}'
        docker_file_res = read_file(docker_file)
        if isinstance(docker_file_res, Bad):
            return Bad(docker_file_res.value)
        docker_file_contents = docker_file_res.value
        all_args = {
            'M_TAG': m_env.m_tag,
            **(extras or {}),
            **self.build_args,
        }
        # Only add the arch if multi-arch is enabled.
        if m_env.multi_arch:
            all_args['ARCH'] = arch
        return Good([
            f'{key}={arg_value}'
            for key, arg_value in all_args.items()
            if f'ARG {key}' in docker_file_contents
        ])

    def format_env_secrets(self: 'DockerImage') -> list[str] | None:
        """Format the environment secrets to pass to the docker build command.

        Returns:
            A list of valid `secret` arguments.
        """
        env_sec = self.env_secrets or []
        return [f'id={env}' for env in env_sec] if env_sec else None

    def ci_build(self: 'DockerImage', m_env: MEnvDocker) -> Res[str]:
        """Generate a shell script to build an image in the CI pipelines.

        Args:
            m_env: The MEnvDocker instance with the environment variables.

        Returns:
            A shell snippet with a docker build command.
        """
        build_args = self.format_build_args(m_env, '"$ARCH"')
        if isinstance(build_args, Bad):
            return Bad(build_args.value)
        build_args.value.append('BUILDKIT_INLINE_CACHE=1')

        docker_file = f'{m_env.base_path}/{self.docker_file}'
        img_name = f'{m_env.registry}/{self.image_name}'
        build_cmd = DockerBuild(
            progress='plain',
            cache_from='staged-image:cache',
            secret=self.format_env_secrets(),
            tag=[
                'staged-image:latest',
                f'{img_name}:{m_env.m_tag}',
            ],
            build_arg=build_args.value,
            target=self.target_stage,
            file=docker_file,
        )
        script = [
            BASH_SHEBANG,
            'export DOCKER_BUILDKIT=1',
            SET_STRICT_BASH,
            '',
            str(build_cmd),
            '',
        ]
        return Good('\n'.join(script))

    def local_build(self: 'DockerImage', m_env: MEnvDocker) -> Res[str]:
        """Generate a shell script to build an image.

        Args:
            m_env: The MEnvDocker instance with the environment variables.

        Returns:
            A shell snippet with a docker build command.
        """
        arch = get_arch()
        build_args = self.format_build_args(m_env, arch)
        if isinstance(build_args, Bad):
            return Bad(build_args.value)

        docker_file = f'{m_env.base_path}/{self.docker_file}'
        img_name = f'{m_env.registry}/{self.image_name}'
        build_cmd = DockerBuild(
            progress='plain',
            secret=self.format_env_secrets(),
            tag=[
                'staged-image:latest',
                f'{img_name}:{m_env.m_tag}',
            ],
            build_arg=build_args.value,
            target=self.target_stage,
            file=docker_file,
        )
        script = [
            BASH_SHEBANG,
            'export DOCKER_BUILDKIT=1',
            SET_STRICT_BASH,
            '',
            str(build_cmd),
            '',
        ]
        return Good('\n'.join(script))



from pydantic import BaseModel


def _to_opt(option: str) -> str:
    opt = option.replace('_', '-', -1)
    if opt.startswith('--'):
        return opt
    return f'--{opt}'


def _to_list(options: dict[str, str | list[str] | bool]) -> list[str]:
    """Convert a dict to a list of strings.

    Args:
        options: The dict to convert.

    Returns:
        A list of strings.
    """
    options_list: list[str] = []
    standard_opts = {_to_opt(key): opt_val for key, opt_val in options.items()}
    for opt_name in sorted(standard_opts):
        opt_val = standard_opts[opt_name]
        if isinstance(opt_val, list):
            for opt_val_item in opt_val:
                options_list.append(f'{opt_name} {opt_val_item}')
        elif isinstance(opt_val, bool) and opt_val:
            options_list.append(opt_name)
        else:
            options_list.append(f'{opt_name} {opt_val}')
    return sorted(options_list)


class ShellCommand(BaseModel):
    """Representation of a shell command."""

    # Name of the program to run. May contain spaces if the program is a subcommand.
    prog: str

    # Arguments to pass to the program.
    # If the key starts with `--` it will be passed as is. Otherwise it will be
    # considered as a python property and will be passed as `--key`.
    options: dict[str, str | list[str] | bool] = {}

    # Positional arguments to pass to the program.
    positional: list[str] = []

    def __str__(self) -> str:
        """Render the command as a string.

        A new line is not added to the end of the string in case the command
        needs to be piped to another command.

        Returns:
            A string representation of the command.
        """
        lines = [self.prog, *_to_list(self.options), *self.positional]
        # We need to join with a backslash and a new line.
        return' \\\n  '.join(lines)  # noqa: WPS342



import math
import os
import sys
from typing import Sequence, TypeVar

from m.color import color
from m.core import Good, OneOf
from m.core.issue import Issue
from m.core.one_of import issue
from m.log.logger import Logger
from pydantic import BaseModel

BaseModelT = TypeVar('BaseModelT', bound=BaseModel)
logger = Logger('m.core.io')


def _is_true(name: str) -> bool:
    return env(name, 'false') in {'true', '1', 'True'}


def is_traceback_enabled() -> bool:
    """Return True if the stacktrace should be displayed.

    This is true by default on a CI environment where the env var CI is set
    to `true`.

    We can enable this locally by settings the env vars

    - DEBUG_M_STACKTRACE
    - DEBUG

    The `DEBUG_M_STACKTRACE` is provided so to only target this particular
    piece of info and avoid other pieces of code that may be looking at
    the `DEBUG` env var from activating.

    Returns:
        True if the stacktrace should be displayed.
    """
    debug_stacktrace = _is_true('DEBUG_M_STACKTRACE')
    debug_mode = _is_true('DEBUG')
    ci_env = _is_true('CI')
    return debug_stacktrace or debug_mode or ci_env


def is_python_info_enabled():
    """Return True if the python file location should be displayed.

    By default this information is not displayed since the stacktrace already
    provides most of the info. If we need it we can set one of the env var

    - DEBUG_M_PYTHON
    - DEBUG

    `DEBUG_M_PYTHON` is accepted in case we do not want to enable the
    traceback.

    Returns:
        True if the python file location should be displayed.
    """
    debug_python = _is_true('DEBUG_M_PYTHON')
    debug_mode = _is_true('DEBUG')
    return debug_python or debug_mode


def format_seconds(number_of_seconds: int | float) -> str:
    """Return a string representing the number of seconds.

    The format is Xd:Xh:Xm:Xs:Xms.

    Args:
        number_of_seconds: The number of seconds to format.

    Returns:
        A friendly representation of the number of seconds.
    """
    milliseconds = int(math.floor(number_of_seconds * 1000))
    milli_sec = milliseconds % 1000
    seconds = int(math.floor(milliseconds / 1000))
    sec = seconds % 60
    minutes = int(math.floor(seconds / 60))
    mins = minutes % 60
    hours = int(math.floor(minutes / 60))
    hrs = hours % 24
    days = int(math.floor(hours / 24))

    entries = [
        f'{days}d' if days else '',
        f'{hrs}h' if hours else '',
        f'{mins}m' if mins else '',
        f'{sec}s' if sec else '',
        f'{milli_sec}ms' if milli_sec else '',
    ]
    return ':'.join([x for x in entries if x]) or '0s'


def env(name: str, def_val: str = '') -> str:
    """Access an environment variable.

    Args:
        name: The name of the environment variable.
        def_val: The default value to return if not set. Defaults to empty.

    Returns:
        The value of the environment variable if set otherwise the `def_val`.
    """
    return os.environ.get(name, def_val)


def renv(key: str) -> OneOf[Issue, str]:
    """Require an environment variable to be defined.

    Args:
        key: The environment variable required to be defined.

    Returns:
        A `OneOf` with the value of the environment variable or an issue.
    """
    env_value = os.environ.get(key)
    # Value may still be an empty string, checking against None
    if env_value is not None:
        return Good(env_value)
    return issue(f'missing {key} in env')


def env_model(model: type[BaseModelT]) -> OneOf[Issue, BaseModelT]:
    """Require multiple env vars to be defined.

    This can be done by defining a model::

        class GithubEnvVars(BaseModel):
            repo: str = Field('GITHUB_REPOSITORY')
            run_id: str = Field('GITHUB_RUN_ID')

    Then we use it::

        print(env.repo for env in env_model(GithubEnvVars)])

    Args:
        model: A pydantic model specifying the environment variables to fetch.

    Returns:
        A `OneOf` with the values of the environment variables or an issue.
    """
    schema = model.model_json_schema()
    missing: list[str] = []
    env_values: dict[str, str] = {}
    for name, field in schema['properties'].items():
        env_name = field.get('default')
        env_value = os.environ.get(env_name)
        if env_value is None:
            missing.append(env_name)
        else:
            env_values[name] = env_value
    if missing:
        missing_str = ', '.join(missing)
        return issue(f'missing [{missing_str}] in env')
    return Good(model(**env_values))


def _ver_str(major: int, minor: int, patch: int) -> str:
    return f'{major}.{minor}.{patch}'


def prompt_next_version(version: str, release_type: str) -> OneOf[Issue, str]:
    """Prompt the developer to select the next version.

    It displays the possible major, minor and patch versions and prompts the
    developer to enter one of them. See https://semver.org/.

    Args:
        version: The current version.
        release_type: If `hotfix` then it bumps the patch value.

    Returns:
        One of the versions wrapped in a OneOf.
    """
    ver = version.split('-')[0]
    parts = [int(x) for x in ver.split('.')]
    patch = _ver_str(parts[0], parts[1], parts[2] + 1)
    if release_type == 'hotfix':
        return Good(patch)

    minor = _ver_str(parts[0], parts[1] + 1, 0)
    major = _ver_str(parts[0] + 1, 0, 0)
    ver = f'{{bold_green}}{version}{{end}}'
    msg = color(f'Current version is {ver}. Enter one of the following:')
    return Good(prompt_choices(msg, [minor, major]))


def prompt_choices(
    description: str,
    choices: Sequence[str],
    as_list=True,
) -> str:
    """Prompt the user with a description until we get one of the choices.

    Args:
        description: The prompt to display.
        choices: List of accepted choices.
        as_list: If false, it will concat the choices with `/`.

    Returns:
        The string chosen by the user.
    """
    valid = False
    dev_input = ''
    choices_display = choices if as_list else '/'.join(choices)
    while not valid:
        logger.prompt(description, {'choices': choices_display})
        sys.stderr.write('>>>>>>>> ')
        dev_input = input()
        if dev_input in choices:
            valid = True
    return dev_input



import sys
from typing import Any, cast

from m.cli import Arg, BaseModel, command
from m.cli.validators import validate_json_payload, validate_payload
from m.core import Bad, one_of


class Arguments(BaseModel):
    """Process a compiler or linter output to determine if the cli should stop.

    examples::

        ~$ m ci celt -t eslint -c @config.json < <(eslint [dir] -f json)

        ~$ eslint [...options] > tmp.json
        ~$ m ci celt -t eslint @tmp.json -c '{"allowedEslintRules":{"semi":1}}'

    Depending on the tool that is chosen the configuration should have an
    entry of the form "allowed[ToolName]Rules" or "ignored[ToolName]Rules".
    Only the first letter of the tool should be capitalized to conform to
    the camel case style.

    The entry should define a map of rule ids to the number of allowed
    violations. In the case of `ignored[ToolName]Rules` we may define
    the rule id and assign an explanation as to why its being ignored.

    In the examples above we use `@config.json`. This means it
    will read the file `config.json`. You can use any file that you want.
    One option is to use eslintrc.json or create a brand new file called
    `allowed_errors.json`.

    Each tool will have different outputs. See below information for each
    tool.

    - eslint: expects json output::

        eslint -f json [dir]

    - pycodestyle: expects default output::

        pycodestyle --format=default [dir]

    - flake8: expects default output::

        flake8 [dir]

    - pylint: expects json output::

        pylint -f json --rcfile=[file] [dir]

    - typescript: expects output::

        tsc --pretty false

    - ruff: expects json output::

        ruff check --format json [dir]
    """

    payload: str = Arg(
        default='@-',
        help='data: @- (stdin), @filename (file), string',
        validator=validate_payload,
        positional=True,
    )

    tool: str = Arg(
        help='name of a supported compiler/linter',
        required=True,
        aliases=['t', 'tool'],
    )

    config: Any = Arg(
        default='{}',  # noqa: P103 - json object, not attempting to format
        aliases=['c', 'config'],
        help='config data: @filename (file), string',
        validator=validate_json_payload,
    )

    max_lines: int = Arg(
        default=5,
        aliases=['m', 'max_lines'],
        help='max number of error lines to print, use -1 for all',
    )

    file_regex: str | None = Arg(
        aliases=['r', 'file_regex'],
        help='regex expression to filter files',
    )

    file_prefix: str | None = Arg(
        aliases=['p', 'file_prefix'],
        help="replace file prefix with 'old1|old2:new'",
    )

    ignore_error_allowance: bool = Arg(
        default=False,
        aliases=['i', 'ignore_error_allowance'],
        help='set every error allowance to 0',
    )

    stats_only: bool = Arg(
        default=False,
        aliases=['s', 'stats_only'],
        help='display a dictionary with current total violations',
    )

    full_message: bool = Arg(
        default=False,
        aliases=['f', 'full_message'],
        help='display the full error message',
    )

    traceback: bool = Arg(
        default=False,
        help='display the exception traceback if available',
    )


@command(
    help='process the output of compiler/linter',
    model=Arguments,
)
def run(arg: Arguments):
    from m.ci.celt.core.process import PostProcessor
    from m.ci.celt.core.types import Configuration
    from m.ci.celt.post_processor import get_post_processor
    from m.core.issue import Issue
    from m.log import Logger

    logger = Logger('m.cli.celt')
    config = Configuration(
        arg.max_lines,
        arg.full_message,
        arg.ignore_error_allowance,
        arg.file_regex,
        arg.file_prefix,
    )
    tool_either = get_post_processor(arg.tool, config)
    either = one_of(lambda: [
        project
        for tool in tool_either
        for project in tool.run(arg.payload, arg.config)
    ])
    if isinstance(either, Bad):
        Issue.show_traceback = arg.traceback
        logger.error_block('celt failure', either.value)
        return 1
    # We can do the check for tool_either but the next cast is valid
    # because either would have failed if tool_either was Bad.
    tool = cast(PostProcessor, tool_either.value)
    project = either.value
    output = (
        tool.stats_json(project)
        if arg.stats_only
        else tool.to_str(project)
    )
    print(output, file=sys.stderr)  # noqa: WPS421
    if project.error_msg:
        logger.error(project.error_msg)
    return project.status.value



from m.core import Bad, Good, Res, hone, io, issue, one_of
from m.github.cli import get_latest_release
from m.log import Logger

from m import git

from .assert_branch import assert_branch
from .release_setup import release_setup
from .release_utils import YES_NO, is_yes

logger = Logger('m.ci.start_release')


def assert_git_status(status: str, description: str) -> Res[bool]:
    """Assert that the current branch is in a clean state.

    This action may stash some changes. This happens when the developer works
    on a hotfix and starts making changes directly in the branch.

    Args:
        status: short key describing the status
        description: The text used to determine the status

    Returns:
        An issue explaining why we cannot complete the release setup. If
        successful it will return a boolean value. A true value means that
        the process stashed changes and they will need to be popped.
    """
    if status == 'clean':
        logger.info('branch is in a clean state')
        return Good(False)
    if status in {'ahead', 'behind'}:
        return issue(
            'branch is not in sync with the remote branch',
            context={
                'git_status': description,
                'suggestion': 'try running `git pull` and/or `git push`',
            },
        )
    can_stash = {
        'Untracked files',
        'Changes to be committed',
        'Changed but not updated',
        'Changes not staged',
    }
    if description in can_stash:
        logger.warning(f'git status: {description}')
        response = io.prompt_choices(
            'would you like to stash the changes and continue?',
            YES_NO,
            as_list=False,
        )
        if is_yes(response):
            return one_of(lambda: [
                True
                for cmd_out in git.stash()
                for _ in logger.info('ran `git stash`', {'git': cmd_out})
            ]).flat_map_bad(hone('git stash failure'))
    return issue(
        'releases can only be done in a clean git state',
        context={
            'git_status': status,
            'description': description,
        },
    )


def verify_release(
    commits: list[str] | None,
    hotfix: bool,
) -> Res[None]:
    """Compare the number of commits to verify if the release should proceed.

    In some cases we may start a hotfix without realizing that there are
    already some commits in the branch that have not been released. In this
    case it should have been a full release instead.

    This step can be skipped if `commits` is `None`.

    Args:
        commits: List of unreleased commits.
        hotfix: flag to let us know if the release is a hotfix.

    Returns:
        None if everything is good, otherwise an issue.
    """
    if commits is None:
        return Good(None)
    if hotfix and len(commits):
        logger.warning('hotfix may contain unreleased features', {
            'commits': commits,
        })
        response = io.prompt_choices(
            'Disregard warning and proceed with hotfix?',
            YES_NO,
            as_list=False,
        )
        if is_yes(response):
            return Good(None)
        return issue('hotfix aborted by user', context={
            'commits': commits,
            'suggestion': 'consider creating a release',
        })
    if not commits and not hotfix:
        logger.warning('there are no commits to release')
        response = io.prompt_choices(
            'Proceed with a release instead of a hotfix?',
            YES_NO,
            as_list=False,
        )
        if is_yes(response):
            return Good(None)
        return issue('release aborted by user', context={
            'commits': 'no commits to release',
            'suggestion': 'consider creating a hotfix',
        })
    return Good(None)


def after_checkout(branch_checkout: str, stashed: bool) -> Res[None]:
    """Notify the user that the branch has switched.

    Optionally if there are stashed changes they will be popped.

    Args:
        branch_checkout: output from git checkout.
        stashed: If True, it will run git stash pop.

    Returns:
        A OneOf with None. There should be no issues. A warning may show up.
    """
    logger.info('branch checkout successful', {
        'git': f'{branch_checkout}\n',
    })
    if stashed:
        pop_result = git.stash_pop()
        if isinstance(pop_result, Bad):
            logger.warning('`git stash pop` issue', pop_result.value)
        else:
            logger.info('stashed files have been restored')
    return Good(None)


def _get_commits(gh_ver: str) -> Res[list[str] | None]:
    either = git.get_commits(gh_ver)
    if isinstance(either, Bad):
        logger.warning(
            'unable to retrieve unreleased commits - skipping checks',
            either.value,
        )
        return Good(None)
    return either


def start_release(gh_token: str, hotfix: bool = False) -> Res[None]:
    """Start the release process.

    Args:
        gh_token: The GITHUB_TOKEN to use to make api calls to Github.
        hotfix: Set to true to start a hotfix.

    Returns:
        None if successful, otherwise an issue.
    """
    release_type = 'hotfix' if hotfix else 'release'
    return one_of(lambda: [
        None
        for config in assert_branch(release_type, 'm')
        for status, description in git.get_status()
        for stashed_changes in assert_git_status(status, description)
        for gh_ver in get_latest_release(gh_token, config.owner, config.repo)
        for commits in _get_commits(gh_ver)
        for _ in verify_release(commits, hotfix=hotfix)
        for new_ver in io.prompt_next_version(gh_ver, release_type)
        for branch_checkout in git.checkout_branch(f'{release_type}/{new_ver}')
        for _ in after_checkout(branch_checkout, stashed_changes)
        for _ in release_setup('m', config, new_ver)
    ])



import json
import os
from typing import Any

from m.core import Bad, Good, Res, issue, one_of, rw
from m.log import Logger
from pydantic import BaseModel

from .env import MEnvDocker
from .filenames import FileNames
from .gh_workflow_multi import Workflow as MultiWorkflow
from .gh_workflow_single import Workflow as SingleWorkflow
from .image import DockerImage
from .shell_scripts import (
    create_cache_script,
    create_push_script,
    create_push_script_tags,
)
from .tags import docker_tags

logger = Logger('m.ci.docker.config')


class DockerConfig(BaseModel):
    """Contains information about the docker images to build."""

    # additional environment variables to inject globally.
    global_env: dict[str, str] | None = None

    # default runner to use when creating blueprints and manifests
    default_runner: str = 'ubuntu-22.04'

    # A map of the architectures to build. It maps say `amd64` to a Github
    # runner that will build the image for that architecture.
    #   amd64: Ubuntu 20.04
    architectures: dict[str, str | list[str]] | None

    # Base path used to locate docker files. Defaults to `.` (root of project)
    # but may be changed a specific directory.
    base_path: str = '.'

    # Name of the docker registry to push the images to. For Github container
    # registry make sure to also include the github owner. For instance:
    # ghcr.io/owner
    docker_registry: str

    # When executing docker build commands we may need to obtain external tokens
    # via other github actions. These can be injected here. We can see those
    # steps in the github workflow file before the actual docker shell scripts
    # are run.
    extra_build_steps: list[dict[str, Any]] | None = None

    # list of images to build
    images: list[DockerImage]

    def makefile_targets(self: 'DockerConfig', files: FileNames) -> str:
        """Create the Makefile targets to trigger the local builds.

        Args:
            files: Instance of FileNames to obtain the names of the scripts.

        Returns:
            A string with the Makefile targets.
        """
        m_dir = files.m_dir
        lines: list[str] = [
            'define m_env',
            f'\t$(eval include {m_dir}/.m/m_env.sh)',
            f'\t$(eval $(cut -d= -f1 {m_dir}/.m/m_env.sh))',
            'endef',
            '',
            'm-env:',
            f'\tmkdir -p {m_dir}/.m && m ci env --bashrc > {m_dir}/.m/m_env.sh',
            '',
            'm-blueprints: m-env',
            '\t$(call m_env)',
            f'\tm blueprints --skip-makefile --skip-workflow {m_dir}',
            f'\tchmod +x {m_dir}/.m/blueprints/local/*.sh\n',
        ]
        for index, img in enumerate(self.images):
            name = img.image_name
            img_file = files.local_file(f'{name}.build.sh')
            previous_img = (
                self.images[index - 1].image_name
                if index > 0
                else None
            )
            dep = f' dev-{previous_img}' if previous_img else ' m-blueprints'
            lines.append(f'dev-{name}:{dep}')
            lines.append('\t$(call m_env)')
            lines.append(f'\t{img_file}\n')
        return '\n'.join(lines)

    def update_makefile(self: 'DockerConfig', files: FileNames) -> Res[int]:
        """Update the Makefile with the docker images targets.

        Args:
            files: Instance of FileNames to obtain the names of the scripts.

        Returns:
            None if successful, else an issue.
        """
        return rw.insert_to_file(
            files.makefile,
            '\n# START: M-DOCKER-IMAGES\n',
            self.makefile_targets(files),
            '\n# END: M-DOCKER-IMAGES\n',
        )

    def update_github_workflow(
        self: 'DockerConfig',
        files: FileNames,
    ) -> Res[int]:
        """Update the github workflow with the docker images targets.

        Args:
            files: Instance of FileNames to obtain the names of the scripts.

        Returns:
            None if successful, else an issue.
        """
        global_env: dict[str, str] = self.global_env or {}
        multi_workflow = MultiWorkflow(
            m_dir=files.m_dir,
            ci_dir=files.ci_dir,
            global_env=global_env,
            default_runner=self.default_runner,
            architectures=self.architectures or {},
            images=self.images,
            extra_build_steps=self.extra_build_steps,
            docker_registry=self.docker_registry,
        )
        single_workflow = SingleWorkflow(
            m_dir=files.m_dir,
            ci_dir=files.ci_dir,
            global_env=global_env,
            default_runner=self.default_runner,
            images=self.images,
            extra_build_steps=self.extra_build_steps,
            docker_registry=self.docker_registry,
        )
        workflow = multi_workflow if self.architectures else single_workflow
        return rw.write_file(files.gh_workflow, str(workflow))

    def write_local_steps(
        self: 'DockerConfig',
        files: FileNames,
        m_env: MEnvDocker,
    ) -> Res[None]:
        """Write local entry point files.

        Args:
            files: The FileNames instance with the file names.
            m_env: The MEnvDocker instance with the environment variables.

        Returns:
            None if successful, else an issue.
        """
        issues: list[dict] = []
        for img in self.images:
            write_res = _write_local_step(files, img, m_env)
            _append_issue(write_res, issues)
        if issues:
            return issue(
                'write_local_steps_failure',
                context={'issues': issues},
            )
        return Good(None)

    def write_ci_steps(
        self: 'DockerConfig',
        files: FileNames,
        m_env: MEnvDocker,
    ) -> Res[None]:
        """Write ci entry point files.

        Args:
            files: The FileNames instance with the file names.
            m_env: The MEnvDocker instance with the environment variables.

        Returns:
            None if successful, else an issue.
        """
        issues: list[dict] = []
        registry = self.docker_registry
        cache_script = create_cache_script(m_env.cache_from_pr, registry)
        push_script = (
            create_push_script(registry)
            if self.architectures
            else create_push_script_tags(registry, m_env.m_tag)
        )
        script_results = [
            rw.write_file(f'{files.ci_dir}/_find-cache.sh', cache_script),
            rw.write_file(f'{files.ci_dir}/_push-image.sh', push_script),
        ]
        for script_res in script_results:
            _append_issue(script_res, issues)
        for img in self.images:
            file_name = f'{files.ci_dir}/{img.image_name}.build.sh'
            write_res = _write_build_script(file_name, img, m_env)
            _append_issue(write_res, issues)
        if issues:
            return issue(
                'write_ci_steps_failure',
                context={'issues': issues},
            )
        return Good(None)

    def write_ci_manifest_info(
        self: 'DockerConfig',
        files: FileNames,
        m_env: MEnvDocker,
    ) -> Res[None]:
        """Write ci entry point files.

        Args:
            files: The FileNames instance with the file names.
            m_env: The MEnvDocker instance with the environment variables.

        Returns:
            None if successful, else an issue.
        """
        if not self.architectures:
            return Good(None)
        m_tag = m_env.m_tag
        if not m_tag and os.environ.get('CI') != 'true':
            logger.warning('M_TAG not found in non-CI environment. Using 1.1.1')
            m_tag = '1.1.1'
        names = [img.image_name for img in self.images]
        tags = [m_tag, *docker_tags(m_tag)]
        names_json = json.dumps(names, separators=(',', ':'))
        tags_json = json.dumps(tags, separators=(',', ':'))
        files_res = [
            rw.write_file(f'{files.ci_dir}/_image-names.json', names_json),
            rw.write_file(f'{files.ci_dir}/_image-tags.json', tags_json),
        ]
        issues: list[dict] = []
        for file_res in files_res:
            _append_issue(file_res, issues)
        if issues:
            return issue(
                'write_ci_manifest_info_failure',
                context={'issues': issues},
            )
        return Good(None)

    def write_blueprints(
        self: 'DockerConfig',
        m_dir: str,
        m_env: MEnvDocker,
    ) -> Res[None]:
        """Write entry point files.

        Writes shell files for both local and ci. Updates the Makefile and
        github workflow file.

        Args:
            m_dir: The m directory.
            m_env: The MEnvDocker instance with the environment variables.

        Returns:
            None if successful, else an issue.
        """
        files = FileNames.create_instance(m_dir)
        return one_of(lambda: [
            None
            for _ in self.write_local_steps(files, m_env)
            for _ in self.write_ci_steps(files, m_env)
            for _ in self.write_ci_manifest_info(files, m_env)
        ])


def _write_build_script(
    file_name: str,
    img: DockerImage,
    m_env: MEnvDocker,
) -> Res[None]:
    return one_of(lambda: [
        None
        for script_content in img.ci_build(m_env)
        for _ in rw.write_file(file_name, script_content)
    ])


def _write_local_step(
    files: FileNames,
    img: DockerImage,
    m_env: MEnvDocker,
) -> Res[None]:
    file_name = files.local_file(f'{img.image_name}.build.sh')
    return one_of(lambda: [
        None
        for script_content in img.local_build(m_env)
        for _ in rw.write_file(file_name, script_content)
    ])


def _append_issue(res: Res[Any], issues: list[dict]):
    # helper function to append issue data to a list
    if isinstance(res, Bad):
        dict_issue = res.value.to_dict(show_traceback=False)
        issues.append(dict(dict_issue))



from m.cli import ArgProxy, BaseModel, command, run_main
from m.core.io import env


class Arguments(BaseModel):
    """Create Github pull request(s) to review the release."""

    github_token: str = ArgProxy(
        '--github-token',
        default=env('GITHUB_TOKEN'),
        help='Github PAT (default: env.GITHUB_TOKEN)',
    )


@command(
    help='create release pull requests',
    model=Arguments,
)
def run(arg: Arguments):
    from m.ci.review_release import review_release

    return run_main(lambda: review_release(arg.github_token))



from m.cli import exec_cli, import_cli_commands
from m.log import logging_config


def main():
    """Execute the cli."""
    logging_config()
    cli_commands = import_cli_commands('m.cli.commands')
    exec_cli(cli_commands)


if __name__ == '__main__':  # pragma: no cover
    main()



from typing import TypeVar

from .fp import OneOf
from .issue import Issue

G = TypeVar('G')  # pylint: disable=invalid-name

# A type alias for a `OneOf` that has an `Issue` on the bad side.
# This is useful for functions that return a `OneOf` that can fail
# with an `Issue`. Shortening for `Result`, `Response` or `Resolution`.
Res = OneOf[Issue, G]



import os

from m.log import Logger

from .tags import docker_tags

logger = Logger('m.ci.docker.shell_scripts')

FIND_CACHE_SCRIPT = """\
#!/bin/bash
imageName=$1
pullCache() {{
  if docker pull -q "$1:$2" 2> /dev/null; then
    docker tag "$1:$2" "staged-image:cache"
  else
    return 1
  fi
}}
findCache() {{
  {find_cache_implementation}
}}
set -euxo pipefail
findCache "{docker_registry}/$imageName"
"""

PUSH_SCRIPT = """\
#!/bin/bash
imageName=$1
set -euxo pipefail
docker tag staged-image:latest "{docker_registry}/$ARCH-$imageName:$M_TAG"
docker push "{docker_registry}/$ARCH-$imageName:$M_TAG"
"""

PUSH_SCRIPT_TAGS = """\
#!/bin/bash
imageName=$1
set -euxo pipefail
{tag_images}
docker image push --all-tags "{docker_registry}/$imageName"
"""


def create_cache_script(pr_num: str, docker_registry: str) -> str:
    """Create a script to retrieve cache for an image.

    Args:
        pr_num: The pull request number where to pull the cache from.
        docker_registry: The docker registry where the images are located.

    Returns:
        A script to find the cache.
    """
    pulls = ['pullCache "$1" master', 'echo "NO CACHE FOUND"']
    if pr_num:
        pulls.insert(0, f'pullCache "$1" "pr{pr_num}"')
    find_cache_implementation = ' || '.join(pulls)
    replacements = {
        'docker_registry': docker_registry,
        'find_cache_implementation': find_cache_implementation,
    }
    return FIND_CACHE_SCRIPT.format(**replacements)


def create_push_script(docker_registry: str) -> str:
    """Create a script to push an image.

    Args:
        docker_registry: The docker registry where the images will be pushed.

    Returns:
        A script to push an image.
    """
    return PUSH_SCRIPT.format(docker_registry=docker_registry)


def create_push_script_tags(docker_registry: str, m_tag: str) -> str:
    """Create a script to push an image.

    This is meant to be used when building for a single architecture.

    Args:
        docker_registry: The docker registry where the images will be pushed.
        m_tag: The unique tag for the image.

    Returns:
        A script to push an image.
    """
    if not m_tag and os.environ.get('CI') != 'true':
        logger.warning('M_TAG not found in non-CI environment. Using 1.1.1')
        m_tag = '1.1.1'
    tags = [m_tag, *docker_tags(m_tag)]
    tagged_images = '\n'.join([
        f'docker tag staged-image:latest "{docker_registry}/$imageName:{tag}"'
        for tag in tags
    ])
    return PUSH_SCRIPT_TAGS.format(
        docker_registry=docker_registry,
        tag_images=tagged_images,
    )



from textwrap import dedent

_indented_snippet = """\
if [ "${CI:-false}" == 'false' ]; then
    alias pnpm="m devcontainer pnpm"
    alias np="m devcontainer pnpm"
    alias cd='HOME=$MDC_WORKSPACE cd'

    function prompter() { export PS1="$(m devcontainer prompter)"; }
    export PROMPT_COMMAND=prompter
fi

export VIRTUAL_ENV="$MDC_VENV_WORKSPACE"
if [ ! -d "$VIRTUAL_ENV" ]; then
    echo "NOTICE: creating virtual environment $VIRTUAL_ENV"
    python3 -m venv "$VIRTUAL_ENV"
fi
. "$VIRTUAL_ENV/bin/activate"
"""

_indented_devex_snippet = """\
    alias pnpm="m devcontainer pnpm"
    alias np="m devcontainer pnpm"
    alias cd='HOME=$MDC_WORKSPACE cd'

    function prompter() { export PS1="$(m devcontainer prompter)"; }
    export PROMPT_COMMAND=prompter
"""

_indented_venv_snippet = """\
    export VIRTUAL_ENV="$MDC_VENV_WORKSPACE"
    if [ ! -d "$VIRTUAL_ENV" ]; then
        echo "NOTICE: creating virtual environment $VIRTUAL_ENV"
        python3 -m venv "$VIRTUAL_ENV"
    fi
    . "$VIRTUAL_ENV/bin/activate"
"""

bashrc_snippet = dedent(_indented_snippet)
devex_snippet = dedent(_indented_devex_snippet)
venv_snippet = dedent(_indented_venv_snippet)



from pydantic import BaseModel


class MEnvDocker(BaseModel):
    """Values from `MEnv` needed by docker."""

    # The m tag to build the images.
    m_tag: str

    # The pull request number to attempt to use as cache.
    cache_from_pr: str

    # base path to locate docker file.
    base_path: str

    # docker registry
    registry: str

    # Flag to indicate if multi-arch is enabled.
    multi_arch: bool



import shlex
import subprocess as sub  # noqa: S404
from subprocess import STDOUT, CalledProcessError  # noqa: S404

from . import Issue, issue
from .fp import Good, OneOf


def eval_cmd(cmd: str) -> OneOf[Issue, str]:
    """Evaluate a bash command and return its output.

    Args:
        cmd: The shell command to evaluate.

    Returns:
        The output of the command (or an Issue if the command failed).
    """
    command = shlex.split(cmd)
    try:
        out = sub.check_output(command, stderr=STDOUT, shell=False).decode()  # noqa: S603,E501
    except CalledProcessError as ex:
        out = ex.output.decode()
        return issue(
            'command returned a non zero exit code',
            context={'cmd': cmd, 'output': out},
        )
    return Good(out.strip())


def exec_pnpm(pnpm_args: list[str]) -> OneOf[Issue, None]:
    """Execute pnpm with the given arguments.

    This command will execute the pnpm command in the current working directory.

    Args:
        pnpm_args: The arguments to pass to pnpm.

    Returns:
        None if successful.
    """
    # delegating the rest of the work to pnpm
    exit_code = sub.call(['pnpm', *pnpm_args], shell=False)  # noqa: S603, S607
    if exit_code:
        return issue('non_zero_pnpm_exit_code', context={
            'pnpm_args': pnpm_args,
            'exit_code': exit_code,
        })
    return Good(None)



from m.cli import BaseModel, command, run_main


class Arguments(BaseModel):
    """Display the very first commit sha in the repository.

    example::

        $ m git first_sha
        bf286e270e13c75dfed289a3921289092477c058
    """


@command(
    help='display the first commit sha',
    model=Arguments,
)
def run():
    from m import git
    return run_main(git.get_first_commit_sha, print)



from enum import Enum


class PullRequestReviewState(str, Enum):  # noqa: WPS600
    """Possible choices for a pull request review state."""

    # A review allowing the pull request to merge.
    approved = 'APPROVED'

    # A review blocking the pull request from merging.
    changes_requested = 'CHANGES_REQUESTED'

    # An informational review.
    commented = 'COMMENTED'

    # A review that has been dismissed.
    dismissed = 'DISMISSED'

    # A review that has not yet been submitted.
    pending = 'PENDING'


class MergeableState(str, Enum):  # noqa: WPS600
    """Whether or not a PullRequest can be merged."""

    # The pull request cannot be merged due to merge conflicts.
    conflicting = 'CONFLICTING'

    # The pull request can be merged.
    mergeable = 'MERGEABLE'

    # The mergeability of the pull request is still being calculated.
    unknown = 'UNKNOWN'



import re


def npm_tags(m_tag: str) -> list[str]:
    """Convert an m_tag to an npm_tag.

    Args:
        m_tag: A tag/version provided by m.

    Returns:
        A tag that can be used by npm during the publishing step.
    """
    tags = []
    regex = r'\d+.\d+.\d+-(.*)\.(.*)'
    matches = re.match(regex, m_tag)
    if matches:
        tag, _ = matches.groups()
        if tag.startswith('rc') or tag.startswith('hotfix'):
            tags.append('next')
        else:
            tags.append(tag)
    else:
        tags.append('latest')
    return tags



from .ci_tools.ci_tools import get_ci_tool
from .ci_tools.types import EnvVars, Message
from .config import logging_config
from .logger import Logger

# using as barrel file to export convenience functions
__all__ = (  # noqa: WPS410
    'get_ci_tool',
    'logging_config',
    'Logger',
    'Message',
    'EnvVars',
)



import argparse

from ..misc import argument_description
from ..types import MISSING, AnyMap, FuncArgs


def should_handle(field: AnyMap) -> bool:
    """Handle the __remainder_args field.

    Args:
        field: A dictionary with information for a cli argument.

    Returns:
        True if it should provide the remainder arguments.
    """
    return field.get('__remainder_args') is True


def handle_field(name: str, field: AnyMap) -> FuncArgs:
    """Set the remainder arguments.

    Args:
        name: The name of the field.
        field: A dictionary representation of the field.

    Returns:
        Function arguments for the parser `add_argument` method.
    """
    args: AnyMap = {
        'help': argument_description(field['description'], MISSING),
        'nargs': argparse.REMAINDER,
    }
    return FuncArgs(args=[name], kwargs=args)



from m.cli import Meta

_desc = """
    CI/CD tools provide ways of interacting with them via shell messages.

    Github Actions:
        https://docs.github.com/en/actions/reference/workflow-commands-for-github-actions

    Teamcity:
        https://www.jetbrains.com/help/teamcity/service-messages.html

    The commands provided here attempt to create an interface to
    display those messages.
"""

meta = Meta(
    help='display messages',
    description=_desc,
)
